from __future__ import annotations

from coredis._utils import EncodingInsensitiveDict, nativestr
from coredis.response._callbacks import ResponseCallback
from coredis.response._callbacks.server import InfoCallback
from coredis.typing import (
    AnyStr,
    ResponsePrimitive,
    ResponseType,
    StringT,
)


def add_flags(result: EncodingInsensitiveDict) -> EncodingInsensitiveDict:
    flags = set(nativestr(result["flags"]).split(","))
    for name, flag in (
        ("is_master", "master"),
        ("is_slave", "slave"),
        ("is_sdown", "s_down"),
        ("is_odown", "o_down"),
        ("is_sentinel", "sentinel"),
        ("is_disconnected", "disconnected"),
        ("is_master_down", "master_down"),
    ):
        result[name] = flag in flags
    return result


class PrimaryCallback(
    ResponseCallback[
        dict[ResponsePrimitive, ResponsePrimitive],
        dict[str, ResponsePrimitive],
    ]
):
    def transform(
        self,
        response: dict[ResponsePrimitive, ResponsePrimitive],
    ) -> dict[str, ResponsePrimitive]:
        return add_flags(EncodingInsensitiveDict(response)).stringify_keys()


class PrimariesCallback(
    ResponseCallback[
        list[dict[StringT, StringT]],
        dict[str, dict[str, ResponsePrimitive]],
    ]
):
    def transform(
        self,
        response: list[dict[StringT, StringT]],
    ) -> dict[str, dict[str, ResponsePrimitive]]:
        states: dict[str, dict[str, ResponsePrimitive]] = {}
        for state in response:
            decoded_state = add_flags(EncodingInsensitiveDict(state)).stringify_keys()
            states[nativestr(decoded_state["name"])] = decoded_state
        return states


class SentinelsStateCallback(
    ResponseCallback[
        list[dict[AnyStr, AnyStr]],
        tuple[dict[str, ResponsePrimitive], ...],
    ]
):
    def transform(
        self,
        response: list[dict[AnyStr, AnyStr]],
    ) -> tuple[dict[str, ResponsePrimitive], ...]:
        return tuple(
            add_flags(EncodingInsensitiveDict(state)).stringify_keys() for state in response
        )


class GetPrimaryCallback(
    ResponseCallback[
        list[AnyStr] | None,
        tuple[str, int] | None,
    ]
):
    def transform(
        self,
        response: list[AnyStr] | None,
    ) -> tuple[str, int] | None:
        if response:
            return nativestr(response[0]), int(response[1])
        return None


class SentinelInfoCallback(
    ResponseCallback[
        list[ResponseType],
        dict[AnyStr, dict[int, dict[str, ResponsePrimitive]]],
    ]
):
    def transform(
        self,
        response: list[ResponseType],
    ) -> dict[AnyStr, dict[int, dict[str, ResponsePrimitive]]]:
        # TODO: giving up for now
        return {response[0]: {r[0]: InfoCallback()(r[1]) for r in response[1]}}  # type: ignore



from __future__ import annotations

from typing import cast

from coredis._json import json
from coredis._utils import nativestr
from coredis.response._callbacks import ResponseCallback
from coredis.response.types import VectorData
from coredis.typing import AnyStr, JsonType, ResponsePrimitive, StringT


class VSimCallback(
    ResponseCallback[
        list[AnyStr]
        | dict[AnyStr, float]
        | dict[AnyStr, AnyStr]
        | dict[AnyStr, list[float | AnyStr]],
        tuple[AnyStr, ...]
        | dict[AnyStr, float]
        | dict[AnyStr, JsonType]
        | dict[AnyStr, tuple[float, JsonType]],
    ],
):
    def transform(
        self,
        response: list[AnyStr]
        | dict[AnyStr, float]
        | dict[AnyStr, AnyStr]
        | dict[AnyStr, list[float | AnyStr]],
    ) -> (
        tuple[AnyStr, ...]
        | dict[AnyStr, float]
        | dict[AnyStr, JsonType]
        | dict[AnyStr, tuple[float, JsonType]]
    ):
        withscores, withattribs = self.options.get("withscores"), self.options.get("withattribs")
        if withscores or withattribs:
            assert isinstance(response, dict)
            match withscores, withattribs:
                case None | False, True:
                    return {
                        k: json.loads(v) for k, v in cast(dict[AnyStr, StringT], response).items()
                    }
                case True, True:
                    return {
                        k: (v[0], json.loads(v[1]))
                        for k, v in cast(dict[AnyStr, tuple[float, StringT]], response).items()
                    }
                case _:
                    return cast(tuple[AnyStr, ...], response)
        else:
            return tuple(response)


class VLinksCallback(
    ResponseCallback[
        list[list[AnyStr]] | list[dict[AnyStr, float]] | None,
        tuple[tuple[AnyStr, ...], ...] | tuple[dict[AnyStr, float], ...] | None,
    ],
):
    def transform(
        self,
        response: list[list[AnyStr]] | list[dict[AnyStr, float]] | None,
    ) -> tuple[tuple[AnyStr, ...], ...] | tuple[dict[AnyStr, float], ...] | None:
        if response:
            if self.options.get("withscores"):
                return tuple(cast(list[dict[AnyStr, float]], response))
            else:
                return tuple(tuple(layer) for layer in cast(list[list[AnyStr]], response))
        return None


class VEmbCallback(
    ResponseCallback[
        list[float] | list[ResponsePrimitive],
        tuple[float, ...] | VectorData | None,
    ]
):
    def transform(
        self,
        response: list[float] | list[ResponsePrimitive] | None,
    ) -> tuple[float, ...] | VectorData | None:
        if response:
            if self.options.get("raw"):
                return VectorData(
                    quantization=nativestr(response[0]),
                    blob=cast(bytes, response[1]),
                    l2_norm=cast(float, response[2]),
                    quantization_range=cast(float, response[3]) if len(response) == 4 else None,
                )
            else:
                return cast(tuple[float, ...], tuple(response))
        return None


class VInfoCallback(
    ResponseCallback[
        dict[AnyStr, AnyStr | int] | None,
        dict[AnyStr, AnyStr | int] | None,
    ]
):
    def transform(
        self,
        response: dict[AnyStr, AnyStr | int] | None,
    ) -> dict[AnyStr, AnyStr | int] | None:
        return response



from __future__ import annotations

import warnings

from .patterns.pipeline import ClusterPipeline, Pipeline

warnings.warn(
    "coredis.pipeline has been moved to coredis.patterns.pipeline",
    DeprecationWarning,
    stacklevel=2,
)
__all__ = ["Pipeline", "ClusterPipeline"]



from __future__ import annotations

from typing import cast

from coredis.modules.response.types import AutocompleteSuggestion
from coredis.response._callbacks import ResponseCallback
from coredis.typing import AnyStr, StringT


class AutocompleteCallback(
    ResponseCallback[
        list[StringT | float | None],
        tuple[AutocompleteSuggestion[AnyStr], ...] | tuple[()],
    ]
):
    def transform(
        self,
        response: list[StringT | float | None],
    ) -> tuple[AutocompleteSuggestion[AnyStr], ...] | tuple[()]:
        if not response:
            return ()
        step = 1
        results = []
        score_idx = payload_idx = 0
        if self.options.get("withscores"):
            score_idx = 1
            step += 1
        if self.options.get("withpayloads"):
            payload_idx = score_idx + 1
            step += 1

        for k in range(0, len(response), step):
            section = response[k : k + step]
            score = section[score_idx] if score_idx else None
            results.append(
                AutocompleteSuggestion[AnyStr](
                    cast(AnyStr, section[0]),
                    float(score) if score else None,
                    cast(AnyStr, section[payload_idx]) if payload_idx else None,
                )
            )

        return tuple(results)



from __future__ import annotations

import inspect
import math
import time
from collections import defaultdict
from contextlib import AsyncExitStack, asynccontextmanager
from typing import TYPE_CHECKING, Any, cast

from anyio import (
    TASK_STATUS_IGNORED,
    AsyncContextManagerMixin,
    CancelScope,
    Event,
    create_memory_object_stream,
    create_task_group,
    move_on_after,
    sleep,
)
from anyio.abc import TaskStatus
from deprecated.sphinx import versionadded, versionchanged

from coredis._utils import b, hash_slot, logger, nativestr
from coredis.commands.constants import CommandName
from coredis.connection import BaseConnection, ClusterConnection, TCPLocation
from coredis.constants.pubsub import (
    PUBLISH_MESSAGE_TYPES,
    SUBSCRIBE_MESSAGE_TYPES,
    SUBUNSUB_MESSAGE_TYPES,
    UNSUBSCRIBE_MESSAGE_TYPES,
    PubSubMessageTypes,
)
from coredis.exceptions import ConnectionError, PubSubError
from coredis.response.types import PubSubMessage
from coredis.retry import (
    CompositeRetryPolicy,
    ConstantRetryPolicy,
    ExponentialBackoffRetryPolicy,
    NoRetryPolicy,
    RetryPolicy,
)
from coredis.typing import (
    AnyStr,
    AsyncGenerator,
    Awaitable,
    Callable,
    Generic,
    Mapping,
    MutableMapping,
    Parameters,
    RedisValueT,
    ResponseType,
    Self,
    StringT,
    TypeVar,
)

if TYPE_CHECKING:
    import coredis.pool

T = TypeVar("T")
PoolT = TypeVar("PoolT", bound="coredis.pool.BaseConnectionPool[Any]")
#: Callables for message handler callbacks. The callbacks
#:  can be sync or async.
SubscriptionCallback = Callable[[PubSubMessage], Awaitable[None]] | Callable[[PubSubMessage], None]


class BasePubSub(AsyncContextManagerMixin, Generic[AnyStr, PoolT]):
    channels: MutableMapping[StringT, SubscriptionCallback | None]
    patterns: MutableMapping[StringT, SubscriptionCallback | None]

    def __init__(
        self,
        connection_pool: PoolT,
        ignore_subscribe_messages: bool = False,
        retry_policy: RetryPolicy | None = CompositeRetryPolicy(
            ExponentialBackoffRetryPolicy(
                (ConnectionError,), retries=None, base_delay=0.1, max_delay=16, jitter=True
            ),
            ConstantRetryPolicy((TimeoutError,), retries=2, delay=0.1),
        ),
        channels: Parameters[StringT] | None = None,
        channel_handlers: Mapping[StringT, SubscriptionCallback] | None = None,
        patterns: Parameters[StringT] | None = None,
        pattern_handlers: Mapping[StringT, SubscriptionCallback] | None = None,
        subscription_timeout: float = 1,
        unsubscription_timeout: float = 0.1,
        max_idle_seconds: float = 15,
    ):
        """
        :param connection_pool: Connection pool used to acquire
         a connection to use for the pubsub consumer
        :param ignore_subscribe_messages: Whether to skip subscription
         acknowledgement messages
        :param retry_policy: An explicit retry policy to use in the subscriber.
        :param channels: channels that the constructed Pubsub instance should
         automatically subscribe to
        :param channel_handlers: Mapping of channels to automatically subscribe to
         and the associated handlers that will be invoked when a message is received
         on the specific channel.
        :param patterns: patterns that the constructed Pubsub instance should
         automatically subscribe to
        :param pattern_handlers: Mapping of patterns to automatically subscribe to
         and the associated handlers that will be invoked when a message is received
         on channel matching the pattern.
        :param subscription_timeout: Maximum amount of time in seconds to wait for
         acknowledgement of subscriptions.
        :param unsubscription_timeout: Maximum amount of time in seconds to wait for
         acknowledgement of unsubscriptions.
        :param max_idle_seconds: Maximum duration (in seconds) to tolerate no
         messages from the server before performing a keepalive check with a
         ``PING``.
        """
        self.connection_pool = connection_pool
        self.ignore_subscribe_messages = ignore_subscribe_messages
        self._connection: coredis.BaseConnection | None = None
        self._retry_policy = retry_policy or NoRetryPolicy()
        self._initial_channel_subscriptions = {
            **{nativestr(channel): None for channel in channels or []},
            **{nativestr(k): v for k, v in (channel_handlers or {}).items()},
        }
        self._initial_pattern_subscriptions = {
            **{nativestr(pattern): None for pattern in patterns or []},
            **{nativestr(k): v for k, v in (pattern_handlers or {}).items()},
        }
        self._send_stream, self._receive_stream = create_memory_object_stream[PubSubMessage | None](
            math.inf
        )
        self._subscribed = Event()
        self._subscription_timeout: float = subscription_timeout
        self._subscription_waiters: dict[StringT, list[Event]] = defaultdict(list)
        self._unsubscription_timeout: float = unsubscription_timeout
        self._unsubscription_waiters: dict[StringT, list[Event]] = defaultdict(list)

        self._last_checkin: float = 0
        self._max_idle_seconds = max_idle_seconds

        self.channels = {}
        self.patterns = {}

    @property
    def connection(self) -> BaseConnection:
        if not self._connection:
            raise Exception("Connection not initialized correctly!")
        return self._connection

    @property
    def subscribed(self) -> bool:
        """Indicates if there are subscriptions to any channels or patterns"""
        return bool(self.channels or self.patterns)

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> PubSubMessage:
        while self._subscribed.is_set():
            if message := await self.get_message():
                return message
        raise StopAsyncIteration()

    @asynccontextmanager
    async def __asynccontextmanager__(self) -> AsyncGenerator[Self]:
        async with (
            self._send_stream,
            self._receive_stream,
            create_task_group() as tg,
        ):
            await tg.start(self.run)
            # initialize subscriptions
            if self._initial_channel_subscriptions:
                await self.subscribe(**self._initial_channel_subscriptions)
            if self._initial_pattern_subscriptions:
                await self.psubscribe(**self._initial_pattern_subscriptions)
            yield self
            self._current_scope.cancel()

    async def run(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        started = False

        async def _run() -> None:
            nonlocal started
            async with self.connection_pool.acquire() as self._connection:
                try:
                    async with create_task_group() as tg:
                        self._current_scope = tg.cancel_scope
                        tg.start_soon(self._consume_push_messages)
                        tg.start_soon(self._keepalive)
                        if not started:
                            task_status.started()
                            started = True
                        else:  # resubscribe
                            if self.channels:
                                await self.subscribe(*self.channels.keys())
                            if self.patterns:
                                await self.psubscribe(*self.patterns.keys())
                finally:
                    if self._current_scope.cancel_called:
                        with CancelScope(shield=True):
                            await self._cleanup()

        await self._retry_policy.call_with_retries(_run)

    async def psubscribe(
        self,
        *patterns: StringT,
        **pattern_handlers: SubscriptionCallback | None,
    ) -> None:
        """
        Subscribes to channel patterns. Patterns supplied as keyword arguments
        expect a pattern name as the key and a callable as the value. A
        pattern's callable will be invoked automatically when a message is
        received on that pattern rather than producing a message via
        :meth:`get_message` or the iterator.
        """
        new_patterns: MutableMapping[StringT, SubscriptionCallback | None] = {}
        new_patterns.update(dict.fromkeys(map(self._encode, patterns)))

        for handled_pattern, handler in pattern_handlers.items():
            new_patterns[self._encode(handled_pattern)] = handler

        waiters: dict[StringT, Event] = {pattern: Event() for pattern in new_patterns}
        for pattern, event in waiters.items():
            self._subscription_waiters[pattern].append(event)

        self._execute_command(CommandName.PSUBSCRIBE, *new_patterns.keys())
        await self._ensure_subscribed(waiters)
        # update the patterns dict AFTER we send the command. we don't want to
        # subscribe twice to these patterns, once for the command and again
        # for the reconnection.
        self.patterns.update(new_patterns)

    async def punsubscribe(self, *patterns: StringT) -> None:
        """
        Unsubscribes from the supplied patterns. If empty, unsubscribe from
        all patterns.
        """
        waiters: dict[StringT, Event] = {
            self._encode(pattern): Event() for pattern in (patterns or self.patterns)
        }
        for pattern, event in waiters.items():
            self._unsubscription_waiters[pattern].append(event)
        self._punsubscribe(*patterns)
        await self._ensure_unsubscribed(waiters)

    async def subscribe(
        self,
        *channels: StringT,
        **channel_handlers: SubscriptionCallback | None,
    ) -> None:
        """
        Subscribes to channels. Channels supplied as keyword arguments expect
        a channel name as the key and a callable as the value. A channel's
        callable will be invoked automatically when a message is received on
        that channel rather than producing a message via :meth:`get_message`
        or the iterator.
        """

        new_channels: MutableMapping[StringT, SubscriptionCallback | None] = {}
        new_channels.update(dict.fromkeys(map(self._encode, channels)))

        for handled_channel, handler in channel_handlers.items():
            new_channels[self._encode(handled_channel)] = handler

        waiters: dict[StringT, Event] = {channel: Event() for channel in new_channels}

        for channel, event in waiters.items():
            self._subscription_waiters[channel].append(event)

        self._execute_command(CommandName.SUBSCRIBE, *new_channels.keys())

        await self._ensure_subscribed(waiters)
        self.channels.update(new_channels)

    async def unsubscribe(self, *channels: StringT) -> None:
        """
        Unsubscribes from the supplied channels. If empty, unsubscribe from
        all channels
        """

        waiters: dict[StringT, Event] = {
            self._encode(channel): Event() for channel in (channels or self.channels)
        }
        for channel, event in waiters.items():
            self._unsubscription_waiters[channel].append(event)
        self._execute_command(CommandName.UNSUBSCRIBE, *channels)
        await self._ensure_unsubscribed(waiters)

    async def get_message(
        self,
        ignore_subscribe_messages: bool = False,
        timeout: int | float | None = None,
    ) -> PubSubMessage | None:
        """
        Gets the next message if one is available, otherwise None.

        :param ignore_subscribe_messages: Whether to skip subscription
         acknowledgement messages
        :param timeout: Number of seconds to wait for a message to be available
         on the connection. If the ``None`` the command will block forever.
        """

        with move_on_after(timeout):
            return self._filter_ignored_messages(
                await self._receive_stream.receive(), ignore_subscribe_messages
            )
        return None

    def _unsubscribe(self, *channels: StringT) -> None:
        self._execute_command(CommandName.UNSUBSCRIBE, *channels)

    def _punsubscribe(self, *patterns: StringT) -> None:
        self._execute_command(CommandName.PUNSUBSCRIBE, *patterns)

    async def _keepalive(self) -> None:
        while True:
            if (idle := time.monotonic() - self._last_checkin) >= self._max_idle_seconds:
                if self._connection and await self._connection.create_request(CommandName.PING) in {
                    b"PONG",
                    "PONG",
                }:
                    self._last_checkin = time.monotonic()
            await sleep(max(1, self._max_idle_seconds - idle))

    async def _consume_push_messages(self) -> None:
        async for response in self.connection.push_messages:
            self._last_checkin = time.monotonic()
            msg = await self._handle_message(response)
            self._send_stream.send_nowait(msg)

    def _encode(self, value: StringT) -> StringT:
        """
        Encodes the value so that it's identical to what we'll read off the
        connection

        :meta private:
        """

        if self.connection_pool.decode_responses and isinstance(value, bytes):
            value = nativestr(value, self.connection_pool.encoding)
        elif not self.connection_pool.decode_responses and isinstance(value, str):
            value = b(value, self.connection_pool.encoding)

        return value

    def _execute_command(self, command: bytes, *args: RedisValueT) -> None:
        """
        Executes a publish/subscribe command

        TODO: evaluate whether this should remain async
        """
        self.connection.send_command(command, *args)

    async def _handle_message(self, response: list[ResponseType]) -> PubSubMessage | None:
        """
        Parses a pub/sub message. If the channel or pattern was subscribed to
        with a message handler, the handler is invoked instead of a parsed
        message being returned.

        :meta private:
        """
        message_type = b(response[0])
        message_type_str = nativestr(response[0])
        message: PubSubMessage

        if message_type in SUBUNSUB_MESSAGE_TYPES:
            target = cast(StringT, response[1])
            message = PubSubMessage(
                type=message_type_str,
                pattern=target if message_type[0] == ord(b"p") else None,
                # This field is populated in all cases for backward compatibility
                # as older versions were incorrectly populating the channel
                # with the pattern on psubscribe/punsubscribe responses.
                channel=target,
                data=cast(int, response[2]),
            )
            if message_type in SUBSCRIBE_MESSAGE_TYPES:
                if waiters := self._subscription_waiters.get(target, []):
                    waiters.pop(-1).set()
            if message_type in UNSUBSCRIBE_MESSAGE_TYPES:
                if waiters := self._unsubscription_waiters.get(target, []):
                    waiters.pop(-1).set()

        elif message_type in PUBLISH_MESSAGE_TYPES:
            if message_type == PubSubMessageTypes.PMESSAGE:
                message = PubSubMessage(
                    type="pmessage",
                    pattern=cast(StringT, response[1]),
                    channel=cast(StringT, response[2]),
                    data=cast(StringT, response[3]),
                )
            else:
                message = PubSubMessage(
                    type="message",
                    pattern=None,
                    channel=cast(StringT, response[1]),
                    data=cast(StringT, response[2]),
                )
        else:
            raise PubSubError(f"Unknown message type {message_type_str}")  # noqa

        # if this is an unsubscribe message, remove it from memory
        if message_type in UNSUBSCRIBE_MESSAGE_TYPES:
            if message_type == PubSubMessageTypes.PUNSUBSCRIBE:
                subscribed_dict = self.patterns
            else:
                subscribed_dict = self.channels
            subscribed_dict.pop(message["channel"], None)

        if message_type in PUBLISH_MESSAGE_TYPES:
            handler = None
            if message_type == PubSubMessageTypes.PMESSAGE and message["pattern"]:
                handler = self.patterns.get(message["pattern"], None)
            elif message["channel"]:
                handler = self.channels.get(message["channel"], None)

            if handler:
                handler_response = handler(message)
                if inspect.isawaitable(handler_response):
                    await handler_response
                return None
        if not (self.channels or self.patterns):
            self._subscribed = Event()

        return message

    def _filter_ignored_messages(
        self,
        message: PubSubMessage | None,
        ignore_subscribe_messages: bool = False,
    ) -> PubSubMessage | None:
        if (
            message
            and b(message["type"]) in SUBUNSUB_MESSAGE_TYPES
            and (self.ignore_subscribe_messages or ignore_subscribe_messages)
        ):
            return None
        return message

    async def _ensure_subscribed(self, waiters: dict[StringT, Event]) -> None:
        with move_on_after(self._subscription_timeout) as cancel_scope:
            for target, event in waiters.items():
                await event.wait()
        if cancel_scope.cancelled_caught:
            raise TimeoutError(
                f"Failed to receive acknowledgement of subscription within {self._subscription_timeout} seconds"
            )

        self._subscribed.set()

    async def _ensure_unsubscribed(self, waiters: dict[StringT, Event]) -> bool:
        with move_on_after(self._unsubscription_timeout) as cancel_scope:
            for _, event in waiters.items():
                await event.wait()
        if cancel_scope.cancelled_caught:
            logger.warning(
                f"Failed to receive acknowledgement of unsubscription within {self._unsubscription_timeout} seconds"
            )
            # TODO: not sure what to do with this response.
            return False
        return True

    async def _cleanup(self) -> None:
        if self.subscribed:
            try:
                self._unsubscribe()
                self._punsubscribe()
                with move_on_after(self._unsubscription_timeout, shield=True):
                    await self._consume_push_messages()
            except ConnectionError:
                return
            finally:
                self._subscription_waiters.clear()
                self._unsubscription_waiters.clear()
                self._subscribed = Event()


@versionchanged(
    version="6.0.0",
    reason="The class supports the async context manager protocol and must always be used as such",
)
class PubSub(BasePubSub[AnyStr, "coredis.pool.ConnectionPool[BaseConnection]"]):
    """
    Pub/Sub implementation to be used with :class:`~coredis.Redis`
    that is returned by :meth:`~coredis.Redis.pubsub`

    An instance of this class is both an async context manager (to
    ensure that proper clean up of connections & subscriptions happens automatically)
    and an async iterator to consume messages from channels or patterns that it is
    subscribed to.

    Recommended use::

        async with coredis.Redis(decode_responses=True) as client:
            async with client.pubsub(
              ignore_subscribe_messages=True,
              channels=["channel-1", "channel-2"]
            ) as pubsub:
                async for message in pubsub:
                    match message["channel"]:
                        case "channel-1":
                            print("first", message["data"])
                        case "channel-2":
                            print("second", message["data"])

    Or to explicitly subscribe::

        async with coredis.Redis(decode_responses=True) as client:
            async with client.pubsub() as pubsub:
                await pubsub.subscribe("channel-1")
                assert (await pubsub.get_message())["channel"] == "channel-1"
                async for message in pubsub:
                    print(message["data"])

    For more details see :ref:`handbook/pubsub:pubsub`
    """


@versionchanged(
    version="6.0.0",
    reason="The class supports the async context manager protocol and must always be used as such",
)
class ClusterPubSub(BasePubSub[AnyStr, "coredis.pool.ClusterConnectionPool"]):
    """
    Pub/Sub implementation to be used with :class:`~coredis.RedisCluster`
    that is returned by :meth:`~coredis.RedisCluster.pubsub`

    .. note:: This implementation does not particularly benefit from having
       multiple nodes in a cluster as it subscribes to messages sent to channels
       using ``PUBLISH`` which in cluster mode results in the message being
       broadcasted to every node in the cluster. For this reason the subscribing
       client can subscribe to any node in the cluster to receive messages sent to
       any channel - which inherently limits the potential for scaling.

       :redis-version:`7.0` introduces the concept of Sharded Pub/Sub which
       can be accessed by instead using :meth:`~coredis.RedisCluster.sharded_pubsub`
       which uses the implementation in :class:`~coredis.commands.ShardedPubSub`.

    An instance of this class is both an async context manager (to
    ensure that proper clean up of connections & subscriptions happens automatically)
    and an async iterator to consume messages from channels or patterns that it is
    subscribed to.

    For more details see :ref:`handbook/pubsub:cluster pub/sub`

    """


@versionchanged(
    version="6.0.0",
    reason="The class supports the async context manager protocol and must always be used as such",
)
@versionadded(version="3.6.0")
class ShardedPubSub(BasePubSub[AnyStr, "coredis.pool.ClusterConnectionPool"]):
    """
    Sharded Pub/Sub implementation to be used with :class:`~coredis.RedisCluster`
    that is returned by :meth:`~coredis.RedisCluster.sharded_pubsub`

    For details about the server architecture refer to the `Redis manual entry
    on Sharded Pub/sub <https://redis.io/docs/develop/pubsub/#sharded-pubsub>`__

    New in :redis-version:`7.0.0`

    .. warning:: Sharded PubSub only supports subscription by channel and does
       **NOT** support pattern based subscriptions.

    An instance of this class is both an async context manager (to
    ensure that proper clean up of connections & subscriptions happens automatically)
    and an async iterator to consume messages from channels that it is subscribed to.

    For more details see :ref:`handbook/pubsub:sharded pub/sub`
    """

    def __init__(
        self,
        connection_pool: coredis.pool.ClusterConnectionPool,
        ignore_subscribe_messages: bool = False,
        retry_policy: RetryPolicy | None = None,
        read_from_replicas: bool = False,
        channels: Parameters[StringT] | None = None,
        channel_handlers: Mapping[StringT, SubscriptionCallback] | None = None,
        subscription_timeout: float = 1,
        unsubscription_timeout: float = 0.1,
        max_idle_seconds: float = 15,
    ):
        """
        :param connection_pool: Connection pool used to acquire
         a connection to use for the pubsub consumer
        :param ignore_subscribe_messages: Whether to skip subscription
         acknowledgement messages
        :param retry_policy: An explicit retry policy to use in the subscriber.
        :param channels: channels that the constructed Pubsub instance should
         automatically subscribe to
        :param channel_handlers: Mapping of channels to automatically subscribe to
         and the associated handlers that will be invoked when a message is received
         on the specific channel.
        :param subscription_timeout: Maximum amount of time in seconds to wait for
         acknowledgement of subscriptions.
        :param unsubscription_timeout: Maximum amount of time in seconds to wait for
         acknowledgement of unsubscriptions.
        :param max_idle_seconds: Maximum duration (in seconds) to tolerate no
         messages from the cluster before performing a keepalive check with a
         `PING``.
        """
        self.shard_connections: dict[str, BaseConnection] = {}
        self._node_channel_mapping: dict[str, list[StringT]] = {}
        self.read_from_replicas = read_from_replicas
        self._last_checkins: dict[TCPLocation, float] = defaultdict(lambda: 0)
        super().__init__(
            connection_pool,
            ignore_subscribe_messages,
            retry_policy,
            channels=channels,
            channel_handlers=channel_handlers,
            subscription_timeout=subscription_timeout,
            unsubscription_timeout=unsubscription_timeout,
            max_idle_seconds=max_idle_seconds,
        )

    async def subscribe(
        self,
        *channels: StringT,
        **channel_handlers: SubscriptionCallback | None,
    ) -> None:
        """
        :param channels: The shard channels to subscribe to.
        :param channel_handlers: Channels supplied as keyword arguments expect
         a channel name as the key and a callable as the value. A channel's
         callable will be invoked automatically when a message is received on
         that channel rather than producing a message via :meth:`get_message`
         or the iterator.
        """

        new_channels: MutableMapping[StringT, SubscriptionCallback | None] = {}
        new_channels.update(dict.fromkeys(map(self._encode, channels)))

        for handled_channel, handler in channel_handlers.items():
            new_channels[self._encode(handled_channel)] = handler

        waiters: dict[StringT, Event] = {channel: Event() for channel in new_channels}

        for channel, event in waiters.items():
            self._subscription_waiters[channel].append(event)
        for channel in new_channels:
            self._execute_command(CommandName.SSUBSCRIBE, channel)
        await self._ensure_subscribed(waiters)
        self.channels.update(new_channels)

    async def unsubscribe(self, *channels: StringT) -> None:
        """
        :param channels: The shard channels to unsubscribe from. If None are provided,
         this will effectively unsubscribe the client from all channels
         previously subscribed to.
        """

        waiters: dict[StringT, Event] = {
            self._encode(channel): Event() for channel in (channels or self.channels)
        }

        for channel, event in waiters.items():
            self._unsubscription_waiters[channel].append(event)
        self._unsubscribe(*(channels or list(self.channels.keys())))
        await self._ensure_unsubscribed(waiters)

    async def psubscribe(
        self,
        *patterns: StringT,
        **pattern_handlers: SubscriptionCallback | None,
    ) -> None:
        """
        Not available in sharded pubsub

        :meta private:
        """
        raise NotImplementedError("Sharded PubSub does not support subscription by pattern")

    async def punsubscribe(self, *patterns: StringT) -> None:
        """
        Not available in sharded pubsub

        :meta private:
        """
        raise NotImplementedError("Sharded PubSub does not support subscription by pattern")

    @asynccontextmanager
    async def __asynccontextmanager__(self) -> AsyncGenerator[Self]:
        async with self._send_stream, self._receive_stream, create_task_group() as tg:
            await tg.start(self.run)
            # initialize subscriptions
            if self._initial_channel_subscriptions:
                await self.subscribe(**self._initial_channel_subscriptions)
            yield self
            self._current_scope.cancel()

    async def run(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        started = False

        async def _run() -> None:
            nonlocal started
            stack = AsyncExitStack()
            self.shard_connections = {
                node.node_id: await stack.enter_async_context(
                    self.connection_pool.acquire(node=node)
                )
                for node in self.connection_pool.cluster_layout.primaries
                if node.node_id
            }
            try:
                async with create_task_group() as tg:
                    self._current_scope = tg.cancel_scope
                    [
                        tg.start_soon(self._consume_shard_push_messages, connection)
                        for connection in self.shard_connections.values()
                    ]
                    [
                        tg.start_soon(self._shard_keepalive, connection)
                        for connection in self.shard_connections.values()
                    ]
                    if not started:
                        task_status.started()
                        started = True
                    elif self.channels:  # resubscribe
                        await self.subscribe(*self.channels.keys())
            finally:
                if self._current_scope.cancel_called:
                    with CancelScope(shield=True):
                        await self._cleanup()

        await self._retry_policy.call_with_retries(_run)

    def _unsubscribe(self, *channels: StringT) -> None:
        for channel in channels or list(self.channels.keys()):
            self._execute_command(CommandName.SUNSUBSCRIBE, channel)

    def _execute_command(self, command: bytes, *args: RedisValueT) -> None:
        assert isinstance(args[0], (bytes, str))
        channel = nativestr(args[0])
        slot = hash_slot(b(channel))
        node = self.connection_pool.cluster_layout.node_for_slot(slot)
        if node and node.node_id:
            key = node.node_id
            self.shard_connections[key].send_command(command, *args)
            return
        raise PubSubError(f"Unable to determine shard for channel {args[0]!r}")

    async def _consume_shard_push_messages(self, connection: BaseConnection) -> None:
        assert isinstance(connection, ClusterConnection)
        async for message in connection.push_messages:
            self._last_checkins[connection.location] = time.monotonic()
            self._send_stream.send_nowait(await self._handle_message(message))

    async def _shard_keepalive(self, connection: BaseConnection) -> None:
        assert isinstance(connection, ClusterConnection)
        while True:
            if (
                idle := time.monotonic() - self._last_checkins[connection.location]
            ) >= self._max_idle_seconds:
                if await connection.create_request(CommandName.PING) in {b"PONG", "PONG"}:
                    self._last_checkins[connection.location] = time.monotonic()
            await sleep(max(1, self._max_idle_seconds - idle))

    async def _cleanup(self) -> None:
        if self.subscribed:
            try:
                self._unsubscribe()
                with move_on_after(self._unsubscription_timeout):
                    async with create_task_group() as tg:
                        for connection in self.shard_connections.values():
                            tg.start_soon(self._consume_shard_push_messages, connection)
            except ConnectionError:
                return
            finally:
                self._subscription_waiters.clear()
                self._unsubscription_waiters.clear()
                self._subscribed = Event()



from __future__ import annotations

from typing import Any, cast

from coredis._utils import EncodingInsensitiveDict
from coredis.response._callbacks import ResponseCallback
from coredis.response._utils import flat_pairs_to_dict
from coredis.response.types import LibraryDefinition
from coredis.typing import (
    AnyStr,
    Mapping,
    MutableMapping,
    RedisValueT,
    ResponsePrimitive,
    ResponseType,
)


class FunctionListCallback(
    ResponseCallback[list[ResponseType], Mapping[AnyStr, LibraryDefinition]]
):
    def transform(
        self,
        response: list[ResponseType],
    ) -> Mapping[AnyStr, LibraryDefinition]:
        libraries = [
            EncodingInsensitiveDict(flat_pairs_to_dict(cast(list[RedisValueT], library)))
            for library in response
        ]
        transformed = EncodingInsensitiveDict()
        for library in libraries:
            lib_name = library["library_name"]
            functions = EncodingInsensitiveDict({})
            for function in library.get("functions", []):
                function_definition = EncodingInsensitiveDict(flat_pairs_to_dict(function))
                functions[function_definition["name"]] = function_definition
                functions[function_definition["name"]]["flags"] = set(function_definition["flags"])
            library["functions"] = functions
            transformed[lib_name] = EncodingInsensitiveDict(
                cast(
                    MutableMapping[str, Any],
                    LibraryDefinition(
                        name=library["library_name"],
                        engine=library["engine"],
                        functions=library["functions"],
                        library_code=library.get("library_code", None),
                    ),
                )
            )
        return transformed


class FunctionStatsCallback(
    ResponseCallback[
        dict[
            AnyStr,
            AnyStr | dict[AnyStr, dict[AnyStr, ResponsePrimitive]] | None,
        ],
        dict[
            AnyStr,
            AnyStr | dict[AnyStr, dict[AnyStr, ResponsePrimitive]] | None,
        ],
    ]
):
    def transform(
        self,
        response: dict[
            AnyStr,
            AnyStr | dict[AnyStr, dict[AnyStr, ResponsePrimitive]] | None,
        ],
    ) -> dict[AnyStr, AnyStr | dict[AnyStr, dict[AnyStr, ResponsePrimitive]] | None]:
        return response



from __future__ import annotations

import functools
import inspect
import itertools
import weakref
from typing import Any, ClassVar, cast, get_args, overload

from deprecated.sphinx import versionadded

from coredis._utils import EncodingInsensitiveDict, nativestr
from coredis.commands.request import CommandRequest, CommandResponseT
from coredis.exceptions import FunctionError
from coredis.response._callbacks import NoopCallback
from coredis.typing import (
    TYPE_CHECKING,
    AnyStr,
    Callable,
    Generator,
    Generic,
    KeyT,
    MutableMapping,
    P,
    Parameters,
    R,
    StringT,
    TypeVar,
    ValueT,
    add_runtime_checks,
    safe_beartype,
)

if TYPE_CHECKING:
    import coredis.client

LibraryT = TypeVar("LibraryT", bound="Library[Any]")
LibraryStringT = TypeVar("LibraryStringT", bound="Library[str]")
LibraryBytesT = TypeVar("LibraryBytesT", bound="Library[bytes]")


class Library(Generic[AnyStr]):
    #: Class variable equivalent of the :paramref:`Library.name` argument.
    NAME: ClassVar[StringT | None] = None
    #: Class variable equivalent of the :paramref:`Library.code` argument.
    CODE: ClassVar[StringT | None] = None

    def __init__(
        self,
        client: coredis.client.Client[AnyStr],
        name: StringT | None = None,
        code: StringT | None = None,
        replace: bool = False,
    ) -> None:
        """
        Abstraction over a library of redis functions

        Example::

            library_code = \"\"\"
            #!lua name=coredis
            redis.register_function('myfunc', function(k, a) return a[1] end)
            \"\"\"
            lib = await Library(client, "mylib", library_code)
            assert "1" == await lib["myfunc"]([], [1])

        When used as a base class the class variables :data:`NAME` and :data:`CODE`
        can be set on the sub class to avoid having to implement a constructor. Constructor
        parameters will take precedence over the class variables.

        :param client: The coredis client instance to use when calling the functions
         exposed by the library.
        :param name: The name of the library (should match the name in the Shebang
         in the library source).
        :param code: The lua code representing the library
        :param replace: Whether to replace the library when intializing. If ``False``
         an exception will be raised if the library was already loaded in the target
         redis instance.
        """
        self._client: weakref.ReferenceType[coredis.client.Client[AnyStr]] = weakref.ref(client)
        self.name = nativestr(name or self.NAME)
        self.code = (code or self.CODE or "").lstrip()
        self._functions: EncodingInsensitiveDict = EncodingInsensitiveDict()
        self.replace = replace
        if self.replace and not self.code:
            raise RuntimeError("library code must be provided when the ``replace`` option is used")

    @property
    def client(self) -> coredis.client.Client[AnyStr]:
        c = self._client()
        assert c
        return c

    @property
    def functions(self) -> MutableMapping[str, Function[AnyStr]]:
        """
        mapping of function names to :class:`~coredis.commands.function.Function`
        instances that can be directly called.
        """
        return self._functions

    async def update(self, new_code: StringT) -> bool:
        """
        Update the code of a library with :paramref:`new_code`
        """
        self.code = new_code
        if await self.initialize(replace=True):
            return True
        return False

    async def initialize(self: LibraryT, replace: bool = False) -> LibraryT:
        from coredis.patterns.pipeline import ClusterPipeline, Pipeline

        self._functions.clear()
        if isinstance(self.client, (Pipeline, ClusterPipeline)):
            redis_client = self.client.client
        else:
            redis_client = self.client
        library = (await redis_client.function_list(self.name)).get(self.name)
        if (not library and self.code) or (replace or self.replace):
            await redis_client.function_load(self.code, replace=replace or self.replace)
            library = (await redis_client.function_list(self.name)).get(self.name)

        if not library:
            raise FunctionError(f"No library found for {self.name}")

        for name, details in library["functions"].items():
            self._functions[name] = Function[AnyStr](
                self.client,
                self.name,
                name,
                bool({b"no-writes", "no-writes"} & details["flags"]),
            )
        return self

    def __await__(self: LibraryT) -> Generator[Any, None, LibraryT]:
        return self.initialize().__await__()

    def __getitem__(self, function: str) -> Function[AnyStr] | None:
        return cast(Function[AnyStr] | None, self._functions.get(function))


@overload
def wraps(
    *,
    function_name: str | None = None,
    runtime_checks: bool = ...,
    readonly: bool = ...,
    verify_existence: bool = ...,
) -> Callable[[Callable[P, CommandRequest[R]]], Callable[P, CommandRequest[R]]]: ...


@overload
def wraps(
    *,
    function_name: str | None = None,
    callback: Callable[..., CommandResponseT],
    runtime_checks: bool = ...,
    readonly: bool = ...,
    verify_existence: bool = ...,
) -> Callable[
    [Callable[P, CommandRequest[Any]]], Callable[P, CommandRequest[CommandResponseT]]
]: ...


@versionadded(version="3.5.0")
def wraps(
    *,
    function_name: str | None = None,
    callback: Callable[..., CommandResponseT] = NoopCallback(),
    runtime_checks: bool = False,
    readonly: bool = False,
    verify_existence: bool = True,
) -> Callable[[Callable[P, CommandRequest[Any]]], Callable[P, CommandRequest[Any]]]:
    """
    Decorator for wrapping methods of subclasses of :class:`Library`
    as entry points to the functions contained in the library. This allows
    exposing a strict signature instead of that which :meth:`Function.__call__`
    provides. The callable being decorated should **not** have an implementation as
    it will never be called. The name of the function decorated must match the foreign
    (Lua) function's name.

    The main objective of the decorator is to allow you to represent a lua library of
    functions as a python class having strict (and type safe) methods as entry points.
    Internally the decorator separates ``keys`` from ``args`` before calling
    :meth:`coredis.Redis.fcall`.

    Mapping the decorated method's arguments to key providers is done by type
    annotations: all parameters annotated as `KeyT` will be passed as keys, and the
    rest will be passed as arguments.

    The following example demonstrates most of the functionality provided by the
    decorator::

        from typing import AnyStr

        import coredis
        from coredis.commands import CommandRequest, Library
        from coredis.commands.function import wraps
        from coredis.typing import KeyT, ValueT

        class MyAwesomeLibrary(Library[AnyStr]):
            NAME = "mylib"
            CODE = \"\"\"
            #!lua name=mylib

            redis.register_function('echo', function(k, a)
              return a[1]
            end)
            redis.register_function('ping', function()
              return "PONG"
            end)
            redis.register_function {
              function_name = 'get',
              callback = function(k, a)
                return redis.call("GET", k[1])
              end,
              flags = { 'no-writes' }  -- mark as read-only
            }
            redis.register_function('hmmget', function(k, a)
              local values = {}
              local fields = {}
              local response = {}
              local i = 1
              local j = 1

              while a[i] do
                fields[j] = a[i]
                i = i + 2
                j = j + 1
              end

              for idx, key in ipairs(k) do
                values = redis.call("HMGET", key, unpack(fields))
                for idx, value in ipairs(values) do
                  if not response[idx] and value then
                    response[idx] = value
                  end
                end
              end
              for idx, value in ipairs(fields) do
                if not response[idx] then
                  response[idx] = a[idx*2]
                end
              end
              return response
            end)
            \"\"\"

            @wraps()
            def echo(self, value: ValueT) -> CommandRequest[ValueT]: ... # type: ignore[empty-body]

            @wraps()
            def ping(self) -> CommandRequest[bytes]: ... # type: ignore[empty-body]

            @wraps(readonly=True)
            def get(self, key: KeyT) -> CommandRequest[ValueT]: ...  # type: ignore[empty-body]

            @wraps()
            def hmmget(self, *keys: KeyT, **fields_with_values: int) -> CommandRequest[list[ValueT]]: ... # type: ignore[empty-body]

        async def run() -> None:
            client = coredis.Redis()
            async with client:
                lib = await MyAwesomeLibrary(client, replace=True)
                await client.set("hello", "world")
                # True
                await lib.echo("hello world")
                # b"hello world"
                await lib.ping()
                # b"PONG"
                await lib.get("hello")
                # b"world"

                async with client.pipeline(transaction=False) as pipe:
                    pipe.hset("k1", {"c": 3, "d": 4})
                    pipe.hset("k2", {"a": 1, "b": 2})
                    lib = await MyAwesomeLibrary(pipe)
                    res = lib.hmmget("k1", "k2", a=-1, b=-2, c=-3, d=-4, e=-5)
                print(await res)
                # [b"1", b"2", b"3", b"4", b"-5"]

    :param function_name: Optional name of the registered library function to map this
     entrypoint to. If not provided, the name will be inferred from the method that the
     decorator wraps
    :param callback: a custom callback to execute on the returned value. When provided,
     the callback's type will be inferred as the return type instead of the type from
     the stub.
    :param runtime_checks: Whether to enable runtime type checking of input arguments
     and return values. (requires :pypi:`beartype`). If :data:`False` the function will
     still get runtime type checking if the environment configuration ``COREDIS_RUNTIME_CHECKS``
     is set - for details see :ref:`handbook/typing:runtime type checking`.
    :param readonly: If ``True`` forces this function to use :meth:`coredis.Redis.fcall_ro`
    :param verify_existence: If ``False``, skip check to see if the function has already
     been registered and call FCALL directly. Use this when you want to save a round trip
     and already know the function is registered.

    :return: A function that has a signature mirroring the decorated function.
    """

    def wrapper(
        func: Callable[P, CommandRequest[Any]],
    ) -> Callable[P, CommandRequest[CommandResponseT]]:
        sig = inspect.signature(func)
        first_arg: str = list(sig.parameters.keys())[0]
        runtime_check_wrapper = add_runtime_checks if not runtime_checks else safe_beartype
        key_params = [
            n
            for n, p in sig.parameters.items()
            if p.annotation == "KeyT" or "KeyT" in get_args(p.annotation)
        ]
        arg_fetch: dict[str, Callable[..., Parameters[Any]]] = {
            n: (
                (lambda v: [v])
                if p.kind
                in {
                    inspect.Parameter.POSITIONAL_ONLY,
                    inspect.Parameter.KEYWORD_ONLY,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                }
                else (
                    (lambda v: list(itertools.chain.from_iterable(v.items())))
                    if p.kind == inspect.Parameter.VAR_KEYWORD
                    else lambda v: list(v)
                )
            )
            for n, p in sig.parameters.items()
        }

        def split_args(
            *a: P.args, **k: P.kwargs
        ) -> tuple[Library[AnyStr], Parameters[KeyT], Parameters[ValueT]]:
            bound_arguments = sig.bind(*a, **k)
            bound_arguments.apply_defaults()
            arguments: dict[str, Any] = bound_arguments.arguments
            instance = arguments.pop(first_arg)
            if not isinstance(instance, Library):
                raise RuntimeError(
                    f"{instance.__class__.__name__} is not a subclass of"
                    " coredis.commands.function.Library therefore it's methods cannot be bound "
                    " to a redis library using ``coredis.commands.function.wrap``."
                    " Please refer to the documentation at https://coredis.readthedocs.org/"
                    " for instructions on how to bind a class to a redis library."
                )
            keys: list[KeyT] = []
            args: list[ValueT] = []
            for name in sig.parameters:
                if name == first_arg:
                    continue
                values = arg_fetch[name](arguments[name])
                if name in key_params:
                    keys.extend(values)
                else:
                    args.extend(values)
            return instance, keys, args

        @runtime_check_wrapper
        @functools.wraps(func)
        def _inner(*args: P.args, **kwargs: P.kwargs) -> CommandRequest[CommandResponseT]:
            name = function_name or func.__name__
            instance, keys, arguments = split_args(*args, **kwargs)
            if (fn := instance.functions.get(name, None)) is None:
                if verify_existence:
                    raise AttributeError(
                        f"Library {instance.name} has no registered function {name}"
                    )
                if readonly:
                    return instance.client.fcall_ro(name, keys, arguments).transform(callback)
                return instance.client.fcall(name, keys, arguments).transform(callback)
            return fn(keys, arguments, readonly=readonly).transform(callback)

        return _inner

    return wrapper


class Function(Generic[AnyStr]):
    def __init__(
        self,
        client: coredis.client.Client[AnyStr],
        library_name: StringT,
        name: StringT,
        readonly: bool = False,
    ):
        """
        Wrapper to call a redis function that has already been loaded

        :param library_name: Name of the library under which the function is registered
        :param name: Name of the function this instance represents
        :param readonly: If ``True`` the function will be called with
         :meth:`coredis.Redis.fcall_ro` instead of :meth:`coredis.Redis.fcall`

        Example::

            func = await Function(client, "mylib", "myfunc")
            response = await func(keys=["a"], args=[1])
        """
        self._client: weakref.ReferenceType[coredis.client.Client[AnyStr]] = weakref.ref(client)
        self.library: Library[AnyStr] = Library[AnyStr](client, library_name)
        self.name = name
        self.readonly = readonly

    @property
    def client(self) -> coredis.client.Client[AnyStr]:
        c = self._client()
        assert c
        return c

    def __call__(
        self,
        keys: Parameters[KeyT] | None = None,
        args: Parameters[ValueT] | None = None,
        callback: Callable[..., CommandResponseT] = NoopCallback(),
        *,
        client: coredis.client.Client[AnyStr] | None = None,
        readonly: bool | None = None,
    ) -> CommandRequest[CommandResponseT]:
        """
        Wrapper to call :meth:`~coredis.Redis.fcall` with the
        function named :paramref:`Function.name` registered under
        the library at :paramref:`Function.library`

        :param keys: The keys this function will reference
        :param args: The arguments expected by the function
        :param callback: a custom callback to call on the raw response from redis before
         returning it.
        :param readonly: If ``True`` forces the function to use :meth:`coredis.Redis.fcall_ro`
        """
        if client is None:
            client = self.client
        if readonly is None:
            readonly = self.readonly

        if readonly:
            return client.fcall_ro(self.name, keys or [], args or []).transform(callback)
        else:
            return client.fcall(self.name, keys or [], args or []).transform(callback)



from __future__ import annotations

import warnings

from .client.sentinel import Sentinel
from .connection._sentinel import SentinelManagedConnection
from .pool._sentinel import SentinelConnectionPool

warnings.warn(
    (
        "The coredis.sentinel submodule has been moved to coredis.{client,pool,connection}."
        " The Sentinel client should be imported from the top level ``coredis`` module."
    ),
    DeprecationWarning,
    stacklevel=2,
)
__all__ = ["Sentinel", "SentinelManagedConnection", "SentinelConnectionPool"]



"""
coredis.commands
----------------
Implementation of core redis commands and abstractions over high level
core concepts such as pubsub, scripting and functions.
"""

from __future__ import annotations

from abc import ABC, abstractmethod

from coredis.response._callbacks import NoopCallback
from coredis.typing import (
    AnyStr,
    Awaitable,
    Callable,
    ExecutionParameters,
    Generic,
    R,
    RedisCommandP,
    Unpack,
    ValueT,
)

# Command wrappers
from .bitfield import BitFieldOperation
from .function import Function, Library
from .request import CommandRequest, CommandResponseT
from .script import Script


class CommandMixin(Generic[AnyStr], ABC):
    @abstractmethod
    def execute_command(
        self,
        command: RedisCommandP,
        callback: Callable[..., R] = NoopCallback(),
        **options: Unpack[ExecutionParameters],
    ) -> Awaitable[R]: ...

    @abstractmethod
    def create_request(
        self,
        name: bytes,
        *arguments: ValueT,
        callback: Callable[..., R],
        execution_parameters: ExecutionParameters | None = None,
    ) -> CommandRequest[R]: ...


__all__ = [
    "CommandRequest",
    "CommandResponseT",
    "BitFieldOperation",
    "Function",
    "Library",
    "Script",
]



from __future__ import annotations

import functools
import hashlib
import inspect
import itertools
from typing import TYPE_CHECKING, Any, cast, get_args, overload

from deprecated.sphinx import versionadded

from coredis._utils import b
from coredis.commands import CommandRequest, CommandResponseT
from coredis.exceptions import NoScriptError
from coredis.response._callbacks import NoopCallback
from coredis.retry import ConstantRetryPolicy
from coredis.typing import (
    AnyStr,
    Awaitable,
    Callable,
    Generic,
    KeyT,
    P,
    Parameters,
    R,
    ResponseType,
    StringT,
    ValueT,
    add_runtime_checks,
    safe_beartype,
)

if TYPE_CHECKING:
    import coredis.client


class Script(Generic[AnyStr]):
    """
    An executable Lua script object returned by :meth:`coredis.Redis.register_script`.
    Instances of the class are callable and take arguments in the same shape
    as :meth:`coredis.Redis.evalsha` or :meth:`coredis.Redis.eval`
    (i.e a list of :paramref:`~__call__.keys` and :paramref:`~__call__.args`).

    Example::

        client = coredis.Redis()
        async with client:
            await client.set("test", "co")
            concat = client.register_script("return redis.call('GET', KEYS[1]) + ARGV[1]")
            assert await concat(['test'], ['redis']) == "coredis"
    """

    #: SHA of this script once it's registered with the redis server
    sha: AnyStr

    def __init__(
        self,
        registered_client: coredis.client.Client[AnyStr] | None = None,
        script: StringT | None = None,
        readonly: bool = False,
    ):
        """
        :param client: The client to use for executing the lua script. If
         this is ``None`` the client will have to be provided when invoking
         the script using :meth:`__call__` with the :paramref:`__call__.client`
         parameter.
        :param script: The lua script that will be used by :meth:`__call__`
        :param readonly: If ``True`` the script will be called with
         :meth:`coredis.Redis.evalsha_ro` instead of :meth:`coredis.Redis.evalsha`
        """
        self.registered_client: coredis.client.Client[AnyStr] | None = registered_client
        self.script: StringT
        if not script:
            raise RuntimeError("No script provided")
        self.script = script
        self.sha = hashlib.sha1(b(script)).hexdigest()  # type: ignore
        self.readonly = readonly

    def __call__(
        self,
        keys: Parameters[KeyT] | None = None,
        args: Parameters[ValueT] | None = None,
        client: coredis.client.Client[AnyStr] | None = None,
        readonly: bool | None = None,
        callback: Callable[..., CommandResponseT] = NoopCallback(),
    ) -> CommandRequest[CommandResponseT]:
        """
        Executes the script registered in :paramref:`Script.script` using
        :meth:`coredis.Redis.evalsha`. Additionally, if the script was not yet
        registered on the instance, it will automatically do that as well
        and cache the sha at :data:`Script.sha`

        :param keys: The keys this script will reference
        :param args: The arguments expected by the script
        :param client: The redis client to use instead of :paramref:`Script.client`
        :param readonly: If ``True`` forces the script to be called with
         :meth:`coredis.Redis.evalsha_ro`
        :param callback: a custom callback to call on the raw response from redis before
         returning it.
        """
        from coredis.patterns.pipeline import ClusterPipeline, Pipeline

        if client is None:
            client = self.registered_client
        if not client:
            raise RuntimeError(
                "This instance is not bound to a redis client."
                "Please provide a valid instance to execute the script with"
            )
        if readonly is None:
            readonly = self.readonly

        method = client.evalsha_ro if readonly else client.evalsha

        if isinstance(client, (ClusterPipeline, Pipeline)):
            cast(Pipeline[AnyStr], client).scripts.add(self)
            return method(self.sha, keys=keys, args=args).transform(callback)
        else:
            return (
                method(self.sha, keys=keys, args=args)
                .retry(
                    ConstantRetryPolicy((NoScriptError,), retries=1, delay=0),
                    lambda _: client.script_load(self.script),
                )
                .transform(callback)
            )

    async def execute(
        self,
        keys: Parameters[KeyT] | None = None,
        args: Parameters[ValueT] | None = None,
        client: coredis.client.Client[AnyStr] | None = None,
        readonly: bool | None = None,
    ) -> ResponseType:
        """
        Executes the script registered in :paramref:`Script.script`

        :meta private:
        """
        return await self(keys, args, client, readonly)

    @overload
    def wraps(
        self,
        *,
        client_arg: str | None = ...,
        runtime_checks: bool = ...,
        readonly: bool = ...,
    ) -> Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]: ...

    @overload
    def wraps(
        self,
        *,
        callback: Callable[..., CommandResponseT],
        client_arg: str | None = ...,
        runtime_checks: bool = ...,
        readonly: bool = ...,
    ) -> Callable[[Callable[P, Awaitable[Any]]], Callable[P, Awaitable[CommandResponseT]]]: ...

    @versionadded(version="3.5.0")
    def wraps(
        self,
        *,
        callback: Callable[..., CommandResponseT] = NoopCallback(),
        client_arg: str | None = None,
        runtime_checks: bool = False,
        readonly: bool = False,
    ) -> Any:
        """
        Decorator for wrapping a regular python function, method or classmethod
        signature with a :class:`~coredis.commands.script.Script`. This allows
        exposing a strict signature instead of that which :meth:`__call__` provides.
        The callable being decorated should **not** have an implementation as
        it will never be called.

        The main objective of the decorator is to allow you to have strict (and type safe)
        signatures for wrappers for lua scripts. Internally the decorator separates
        ``keys`` from ``args`` before calling :meth:`coredis.Redis.evalsha`.

        Mapping the decorated method's arguments to key providers is done by type
        annotations: all parameters annotated as `KeyT` will be passed as keys, and the
        rest will be passed as arguments.

        By default the decorated method is bound to the :class:`coredis.client.Redis`
        or :class:`coredis.client.RedisCluster` instance that the :class:`Script` instance
        was instantiated with. This may however not be the instance you want to eventually
        execute the method with. For such scenarios the decorated method can accept an additional
        parameter which has the name declared by :paramref:`client_arg`).

        The following example executes the script with the ``client`` instance used
        to register the script. The ``key`` parameter is detected as a key provider
        as it is annotated with the :data:`coredis.typing.KeyT` type, and ``value`` is
        passed to redis as an ``arg``::

            import coredis
            from coredis.typing import KeyT, ValueT
            from coredis.commands import Command Request

            client = coredis.Redis()
            @client.register_script("return {KEYS[1], ARGV[1]}").wraps()
            def echo_key_value(key: KeyT, value: ValueT) -> CommandRequest[list[ValueT]]: ...

            async with client:
                res = await echo_key_value("co", "redis")
            # [b"co", b"redis"]

        Alternatively, the following example builds a class method that requires
        the ``client`` to be passed in explicitly::

            from coredis import Redis
            from coredis.commands import Script
            from coredis.typing import StringT

            class ScriptProvider:
                @classmethod
                @Script(script="return KEYS[1]").wraps(
                    client_arg="client"
                )
                def echo_key(cls, client, key: KeyT) -> CommandRequest[StringT]: ...

                @classmethod
                @Script(script="return ARGS[1]").wraps(
                    client_arg="client"
                )
                def echo_arg(cls, client, value) -> CommandRequest[StringT]: ...

            async with Redis() as client:
                echoed = await ScriptProvider.echo_key(client, "coredis")
                # b"coredis"
                echoed = await ScriptProvider.echo_value(client, "coredis")
                # b"coredis"

        You can also pass a custom callback to execute on the return type, which will
        be inferred as the return type rather than the annotation::

            class MyCallback(ResponseCallback[Any, Any, int]):
                def transform(self, response: ResponseType) -> int:
                    return sum([ord(c) for c in str(response)])

            client = coredis.Redis(decode_responses=True)
            async with client:
                script = client.register_script("return {KEYS[1], ARGV[1]}")

                # we use Any since return type will come from callback
                @script.wraps(callback=MyCallback())
                def echo_key_value(key: KeyT, value: ValueT) -> CommandRequest[int]: ...

                res = await echo_key_value("co", "redis")
                reveal_type(res)  # int
                # 1161

        :param callback: a custom callback to execute on the returned value. When provided,
         the callback's type will be inferred as the return type instead of the type from
         the stub.
        :param client_arg: The parameter of the decorator that will contain a client instance
         to be used to execute the script.
        :param runtime_checks: Whether to enable runtime type checking of input arguments
         and return values. (requires :pypi:`beartype`). If :data:`False` the function will
         still get runtime type checking if the environment configuration ``COREDIS_RUNTIME_CHECKS``
         is set - for details see :ref:`handbook/typing:runtime type checking`.
        :param readonly: If ``True`` forces this script to be called with
         :meth:`coredis.Redis.evalsha_ro`

        :return: A function that has a signature mirroring the decorated function.
        """

        def wrapper(func: Callable[P, Awaitable[Any]]) -> Callable[P, Awaitable[CommandResponseT]]:
            sig = inspect.signature(func)
            first_arg = None
            if args := list(sig.parameters.keys()):
                first_arg = args[0]
            runtime_check_wrapper = add_runtime_checks if not runtime_checks else safe_beartype
            script_instance = self
            key_params = [
                n
                for n, p in sig.parameters.items()
                if p.annotation == "KeyT" or "KeyT" in get_args(p.annotation)
            ]
            arg_fetch: dict[str, Callable[..., Parameters[Any]]] = {
                n: (
                    (lambda v: [v])
                    if p.kind
                    in {
                        inspect.Parameter.POSITIONAL_ONLY,
                        inspect.Parameter.KEYWORD_ONLY,
                        inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    }
                    else (
                        (lambda v: list(itertools.chain.from_iterable(v.items())))
                        if p.kind == inspect.Parameter.VAR_KEYWORD
                        else lambda v: list(v)
                    )
                )
                for n, p in sig.parameters.items()
            }
            if first_arg in {"self", "cls"}:
                arg_fetch.pop(first_arg)
            if client_arg:
                arg_fetch.pop(client_arg)

            def split_args(
                bound_arguments: inspect.BoundArguments,
            ) -> tuple[
                Parameters[KeyT],
                Parameters[ValueT],
                coredis.client.Client[AnyStr] | None,
            ]:
                bound_arguments.apply_defaults()
                arguments = bound_arguments.arguments
                keys: list[KeyT] = []
                args: list[ValueT] = []
                for name in sig.parameters:
                    if name not in arg_fetch:
                        continue
                    value = arg_fetch[name](arguments[name])
                    if name in key_params:
                        keys.extend(value)
                    else:
                        args.extend(value)
                return (
                    keys,
                    args,
                    arguments.get(client_arg) if client_arg else None,
                )

            @runtime_check_wrapper
            @functools.wraps(func)
            def __inner(
                *args: P.args,
                **kwargs: P.kwargs,
            ) -> CommandRequest[CommandResponseT]:
                keys, arguments, client = split_args(sig.bind(*args, **kwargs))
                return script_instance(keys, arguments, client, readonly, callback=callback)

            return __inner

        return wrapper



from __future__ import annotations

from typing import cast

from coredis.exceptions import StalePrimaryError
from coredis.typing import ResponseType, Unpack

from ._base import BaseConnectionParams
from ._tcp import TCPConnection, TCPLocation


class SentinelManagedConnection(TCPConnection):
    def __init__(
        self,
        location: TCPLocation,
        primary_name: str,
        is_primary: bool,
        **kwargs: Unpack[BaseConnectionParams],
    ):
        self.primary_name = primary_name
        self.is_primary = is_primary
        super().__init__(location=location, **kwargs)
        if self.is_primary:
            self.register_connect_callback(self._verify_primary_role)

    async def _verify_primary_role(self, connection: SentinelManagedConnection) -> None:
        role = cast(list[ResponseType], await connection.create_request(b"ROLE", decode=True))
        if not role[0] == "master":
            raise StalePrimaryError(
                f"Primary {self.primary_name} at {self.location} reports a role of {role[0]!r}"
            )

    def __repr__(self) -> str:
        host_info = f",host={self.location.host},port={self.location.port}"
        return f"{type(self).__name__}<service={self.primary_name}{host_info}>"



from __future__ import annotations

from typing import Any

from coredis._utils import EncodingInsensitiveDict
from coredis.response._callbacks import ResponseCallback, SimpleStringCallback
from coredis.response.types import LCSMatch, LCSResult
from coredis.typing import (
    AnyStr,
    ResponsePrimitive,
    ResponseType,
)


class StringSetCallback(ResponseCallback[AnyStr | None, AnyStr | bool | None]):
    def transform(self, response: AnyStr | None, **options: Any) -> AnyStr | bool | None:
        if self.options.get("get"):
            return response
        else:
            return SimpleStringCallback()(response)


class LCSCallback(
    ResponseCallback[
        dict[ResponsePrimitive, ResponseType],
        LCSResult,
    ]
):
    def transform(
        self,
        response: dict[ResponsePrimitive, ResponseType],
        **options: Any,
    ) -> LCSResult:
        proxy = EncodingInsensitiveDict(response)

        return LCSResult(
            tuple(
                LCSMatch(
                    (int(k[0][0]), int(k[0][1])),
                    (int(k[1][0]), int(k[1][1])),
                    k[2] if len(k) > 2 else None,
                )
                for k in proxy["matches"]
            ),
            proxy["len"],
        )



from __future__ import annotations

import dataclasses
import math
import random
import time
from abc import ABC, abstractmethod
from functools import wraps
from typing import Any

from anyio import sleep
from exceptiongroup import BaseExceptionGroup

from coredis._utils import logger
from coredis.typing import Awaitable, Callable, Generator, P, R


@dataclasses.dataclass
class Attempt:
    attempt: int
    final: bool


class RetryPolicy(ABC):
    """
    Abstract retry policy
    """

    def __init__(
        self,
        retryable_exceptions: tuple[type[BaseException], ...],
        retries: int | None,
        *,
        deadline: float = math.inf,
    ) -> None:
        """
        :param retryable_exceptions: The exceptions to trigger a retry for
        :param retries: number of times to retry if a :paramref:`retryable_exception`
         is encountered.
        :param deadline: Stop retrying when the time from the first attempt > deadline

        .. warning:: If :paramref:`retries` is ``None`` and deadline is ``math.inf``
           this policy effectively becomes an infinite retry policy.
        """
        self.retryable_exceptions = retryable_exceptions
        self.__retries = retries
        self.__deadline = deadline

    @abstractmethod
    def delay(self, attempt_number: int) -> float:
        """
        Returns the amount of time to pause after the ``attempt_number`` attempt
        """
        raise NotImplementedError()

    def attempts(self) -> Generator[Attempt]:
        attempt = 1
        start = time.monotonic()
        retries_complete = False
        while True:
            now = time.monotonic()
            if retries_complete:
                break
            retries_complete = (self.__retries is not None and attempt == self.__retries + 1) or (
                now + self.delay(attempt) - start >= self.__deadline
            )
            yield Attempt(attempt, retries_complete)
            attempt += 1

    async def call_with_retries(
        self,
        func: Callable[..., Awaitable[R]],
        before_hook: Callable[..., Awaitable[Any]] | None = None,
        failure_hook: Callable[..., Awaitable[Any]]
        | dict[type[BaseException], Callable[..., Awaitable[None]]]
        | None = None,
    ) -> R:
        """
        :param func: a function that should return the coroutine that will be
         awaited when retrying if :paramref:`RetryPolicy.retryable_exceptions` is encountered.
        :param before_hook: if provided will be called on every attempt.
        :param failure_hook: if provided and is a callable it will be
         called everytime a retryable exception is encountered. If it is a mapping
         of exception types to callables, the first exception type that is a parent
         of any encountered exception will be called.
        """
        last_error: BaseException | None = None
        for attempt in self.attempts():
            try:
                if before_hook:
                    await before_hook()
                return await func()
            except BaseException as e:
                if self.will_retry(e):
                    logger.info(f"Retry attempt {attempt.attempt} due to error: {e}")
                    if failure_hook:
                        try:
                            if isinstance(failure_hook, dict):
                                for exc_type, hook in failure_hook.items():
                                    if self._exception_matches(e, exc_type):
                                        await hook(e)
                                        break
                            else:
                                await failure_hook(e)
                        except:  # noqa
                            pass
                    last_error = e
                    if not attempt.final:
                        await sleep(self.delay(attempt.attempt))
                else:
                    raise
        assert last_error
        raise last_error

    def will_retry(self, exc: BaseException) -> bool:
        return RetryPolicy._exception_matches(exc, *self.retryable_exceptions)

    @classmethod
    def _exception_matches(cls, needle: BaseException, *haystack: type[BaseException]) -> bool:
        if isinstance(needle, BaseExceptionGroup):
            for exc in haystack:
                match, unmatched = needle.split(exc)
                if match:
                    return True
        else:
            return isinstance(needle, haystack)
        return False

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}<"
            f"retries={self.__retries}, "
            f"deadline={self.__deadline}, "
            f"retryable_exceptions={','.join(e.__name__ for e in self.retryable_exceptions)}"
            ">"
        )


class NoRetryPolicy(RetryPolicy):
    def __init__(self) -> None:
        super().__init__(retryable_exceptions=(), retries=0, deadline=0)

    def delay(self, attempt_number: int) -> float:
        return 0


class ConstantRetryPolicy(RetryPolicy):
    """
    Retry policy that pauses :paramref:`delay`
    seconds between :paramref:`retries` attempts
    or until :paramref:`deadline` is met if any of
    :paramref:`retryable_exceptions` are encountered.
    """

    def __init__(
        self,
        retryable_exceptions: tuple[type[BaseException], ...],
        retries: int | None,
        *,
        deadline: float = math.inf,
        delay: float = 1,
    ) -> None:
        self.__delay = delay
        super().__init__(
            retryable_exceptions=retryable_exceptions, retries=retries, deadline=deadline
        )

    def delay(self, attempt_number: int) -> float:
        return self.__delay


class ExponentialBackoffRetryPolicy(RetryPolicy):
    """
    Retry policy that exponentially backs off before retrying up to
    :paramref:`retries` or :paramref:`deadline` if any
    of :paramref:`retryable_exceptions` are encountered.
    :paramref:`base_delay` is used as the base value for calculating the
    exponential backoff given the attempt.

    For example with ``base_delay`` == 1::

        attempt 1 = 2^(1-1)*1 == 1
        attempt 2 = 2^(2-1)*1 == 2
        attempt 3 = 2^(3-1)*1 == 4

    To cap the delay to a maximum value, use :paramref:`max_delay`.

    If :paramref:`jitter` is true the delay will be randomly varied
    between :paramref:`base_delay` and the exponential delay for the
    given attempt
    """

    def __init__(
        self,
        retryable_exceptions: tuple[type[BaseException], ...],
        retries: int | None,
        *,
        deadline: float = math.inf,
        base_delay: float = 1,
        max_delay: float = math.inf,
        jitter: bool = False,
    ) -> None:
        self.__base_delay = base_delay
        self.__max_delay = max_delay
        self.__jitter = jitter
        super().__init__(
            retryable_exceptions=retryable_exceptions, retries=retries, deadline=deadline
        )

    def delay(self, attempt_number: int) -> float:
        delay: float = min(self.__max_delay, pow(2, attempt_number - 1) * self.__base_delay)
        if self.__jitter:
            delay = random.uniform(self.__base_delay, delay)
        return delay


class CompositeRetryPolicy(RetryPolicy):
    """
    Convenience class to combine multiple retry policies
    """

    def __init__(self, *retry_policies: RetryPolicy):
        self._retry_policies = set(retry_policies)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{','.join(str(p) for p in self._retry_policies)}>"

    def add_retry_policy(self, policy: RetryPolicy) -> None:
        """
        Add to the retry policies that this instance was created with
        """
        self._retry_policies.add(policy)

    def delay(self, attempt_number: int) -> float:
        raise NotImplementedError()

    async def call_with_retries(
        self,
        func: Callable[..., Awaitable[R]],
        before_hook: Callable[..., Awaitable[Any]] | None = None,
        failure_hook: None
        | (
            Callable[..., Awaitable[Any]] | dict[type[BaseException], Callable[..., Awaitable[Any]]]
        ) = None,
    ) -> R:
        """
        Calls :paramref:`func` repeatedly according to the retry policies that
        this class was instantiated with (:paramref:`CompositeRetryPolicy.retry_policies`).

        :param func: a function that should return the coroutine that will be
         awaited when retrying if :paramref:`RetryPolicy.retryable_exceptions` is encountered.
        :param before_hook: if provided will be called before every attempt.
        :param failure_hook: if provided and is a callable it will be
         called everytime a retryable exception is encountered. If it is a mapping
         of exception types to callables, the first exception type that is a parent
         of any encountered exception will be called.
        """

        policy_attempts = {policy: policy.attempts() for policy in self._retry_policies}
        total_attempts: int = 0
        while True:
            try:
                total_attempts += 1
                if before_hook:
                    await before_hook()
                return await func()
            except BaseException as e:
                retry_delays = []
                for policy, attempts in policy_attempts.items():
                    if policy.will_retry(e):
                        try:
                            attempt = next(attempts)
                            if not attempt.final:
                                retry_delays.append(policy.delay(attempt.attempt))
                        except StopIteration:
                            pass  # This policy is exhausted

                if failure_hook:
                    if isinstance(failure_hook, dict):
                        for exc_type, hook in failure_hook.items():
                            if RetryPolicy._exception_matches(e, exc_type):
                                await hook(e)
                                break
                    else:
                        await failure_hook(e)

                if retry_delays:
                    logger.info(f"Retry attempt {total_attempts} due to error: {e}")
                    await sleep(max(retry_delays))
                    continue
                raise e


def retryable(
    policy: RetryPolicy,
    failure_hook: Callable[..., Awaitable[Any]] | None = None,
) -> Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]:
    """
    Decorator to be used to apply a retry policy to a coroutine
    """

    def inner(
        func: Callable[P, Awaitable[R]],
    ) -> Callable[P, Awaitable[R]]:
        @wraps(func)
        async def _inner(*args: P.args, **kwargs: P.kwargs) -> R:
            return await policy.call_with_retries(
                lambda: func(*args, **kwargs), failure_hook=failure_hook
            )

        return _inner

    return inner



from __future__ import annotations

import functools
import inspect

from coredis.config import Config
from coredis.exceptions import CommandSyntaxError
from coredis.typing import (
    Callable,
    Iterable,
    ParamSpec,
    TypeVar,
)

R = TypeVar("R")
P = ParamSpec("P")


class RequiredParameterError(CommandSyntaxError):
    def __init__(self, arguments: set[str], details: str | None):
        message = (
            f"One of [{','.join(arguments)}] must be provided.{' ' + details if details else ''}"
        )
        super().__init__(arguments, message)


class MutuallyExclusiveParametersError(CommandSyntaxError):
    def __init__(self, arguments: set[str], details: str | None):
        message = (
            f"The [{','.join(arguments)}] parameters are mutually exclusive."
            f"{' ' + details if details else ''}"
        )
        super().__init__(arguments, message)


class MutuallyInclusiveParametersMissing(CommandSyntaxError):
    def __init__(self, arguments: set[str], leaders: set[str], details: str | None):
        if leaders:
            message = (
                f"The [{','.join(arguments)}] parameters(s)"
                f" must be provided together with [{','.join(leaders)}]."
                f"{' ' + details if details else ''}"
            )
        else:
            message = (
                f"The [{','.join(arguments)}] parameters are mutually"
                " inclusive and must be provided together."
                f"{' ' + details if details else ''}"
            )
        super().__init__(arguments, message)


def mutually_exclusive_parameters(
    *exclusive_params: str | Iterable[str],
    details: str | None = None,
    required: bool = False,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    primary = {k for k in exclusive_params if isinstance(k, str)}
    secondary = [k for k in set(exclusive_params) - primary]

    def wrapper(
        func: Callable[P, R],
    ) -> Callable[P, R]:
        sig = inspect.signature(func)

        @functools.wraps(func)
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            if not Config.optimized:
                call_args = sig.bind_partial(*args, **kwargs)
                params = {
                    k
                    for k in primary
                    if not call_args.arguments.get(k) == getattr(sig.parameters.get(k), "default")
                }

                if params:
                    for group in secondary:
                        for k in group:
                            if not call_args.arguments.get(k) == getattr(
                                sig.parameters.get(k), "default"
                            ):
                                params.add(k)

                                break

                if len(params) > 1:
                    raise MutuallyExclusiveParametersError(params, details)
                if len(params) == 0 and required:
                    raise RequiredParameterError(primary, details)

            return func(*args, **kwargs)

        return wrapped if not Config.optimized else func

    return wrapper


def mutually_inclusive_parameters(
    *inclusive_params: str,
    leaders: Iterable[str] | None = None,
    details: str | None = None,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    _leaders = set(leaders or [])
    _inclusive_params = set(inclusive_params)

    def wrapper(
        func: Callable[P, R],
    ) -> Callable[P, R]:
        sig = inspect.signature(func)

        @functools.wraps(func)
        def wrapped(*args: P.args, **kwargs: P.kwargs) -> R:
            if not Config.optimized:
                call_args = sig.bind_partial(*args, **kwargs)
                params = {
                    k
                    for k in _inclusive_params | _leaders
                    if not call_args.arguments.get(k) == getattr(sig.parameters.get(k), "default")
                }
                if _leaders and _leaders & params != _leaders and len(params) > 0:
                    raise MutuallyInclusiveParametersMissing(_inclusive_params, _leaders, details)
                elif not _leaders and params and len(params) != len(_inclusive_params):
                    raise MutuallyInclusiveParametersMissing(_inclusive_params, _leaders, details)
            return func(*args, **kwargs)

        return wrapped if not Config.optimized else func

    return wrapper



from __future__ import annotations

import contextlib
import contextvars
import random
import warnings
from ssl import SSLContext
from typing import TYPE_CHECKING, Any, cast, overload

from anyio import AsyncContextManagerMixin, sleep
from deprecated.sphinx import versionadded
from exceptiongroup import catch
from packaging.version import InvalidVersion, Version

from coredis._utils import logger, nativestr
from coredis.commands import CommandRequest
from coredis.commands._key_spec import KeySpec
from coredis.commands._validators import (
    mutually_inclusive_parameters,
)
from coredis.commands.constants import CommandFlag, CommandName
from coredis.commands.core import CoreCommands
from coredis.commands.function import Library
from coredis.commands.script import Script
from coredis.commands.sentinel import SentinelCommands
from coredis.config import Config
from coredis.connection._base import (
    BaseConnection,
    Location,
    RedisSSLContext,
)
from coredis.connection._tcp import TCPConnection, TCPLocation
from coredis.connection._uds import UnixDomainSocketConnection, UnixDomainSocketLocation
from coredis.credentials import AbstractCredentialProvider
from coredis.exceptions import (
    ConnectionError,
    PersistenceError,
    ReplicationError,
    WatchError,
)
from coredis.globals import CACHEABLE_COMMANDS, COMMAND_FLAGS, READONLY_COMMANDS
from coredis.modules import ModuleMixin
from coredis.patterns.cache import AbstractCache
from coredis.patterns.pubsub import PubSub, SubscriptionCallback
from coredis.pool._base import BaseConnectionPool
from coredis.pool._basic import ConnectionPool, ConnectionPoolParams
from coredis.response._callbacks import (
    NoopCallback,
)
from coredis.response.types import ScoredMember
from coredis.retry import (
    CompositeRetryPolicy,
    ConstantRetryPolicy,
    ExponentialBackoffRetryPolicy,
    NoRetryPolicy,
    RetryPolicy,
)
from coredis.typing import (
    AnyStr,
    AsyncGenerator,
    AsyncIterator,
    Awaitable,
    Callable,
    ExecutionParameters,
    Generic,
    Iterator,
    KeyT,
    Literal,
    Mapping,
    Parameters,
    ParamSpec,
    RedisCommandP,
    RedisValueT,
    Self,
    StringT,
    T_co,
    TypeAdapter,
    TypeVar,
    Unpack,
    ValueT,
)

P = ParamSpec("P")
R = TypeVar("R")

if TYPE_CHECKING:
    import coredis.patterns.pipeline
    from coredis.patterns.lock import Lock
    from coredis.patterns.streams import Consumer, GroupConsumer, StreamParameters

ClientT = TypeVar("ClientT", bound="Client[Any]")
RedisT = TypeVar("RedisT", bound="Redis[Any]")


class Client(
    AsyncContextManagerMixin,
    Generic[AnyStr],
    CoreCommands[AnyStr],
    ModuleMixin[AnyStr],
    SentinelCommands[AnyStr],
):
    connection_pool: BaseConnectionPool[Any]
    decode_responses: bool
    encoding: str
    server_version: Version | None
    type_adapter: TypeAdapter

    def __init__(
        self,
        connection_pool: BaseConnectionPool[Any],
        verify_version: bool = True,
        noreply: bool = False,
        retry_policy: RetryPolicy = NoRetryPolicy(),
        type_adapter: TypeAdapter | None = None,
    ):

        self.connection_pool = connection_pool
        self.encoding = connection_pool.encoding
        self.decode_responses = connection_pool.decode_responses
        self.server_version: Version | None = None
        self.verify_version = verify_version
        self.__noreply = noreply
        self._noreplycontext: contextvars.ContextVar[bool | None] = contextvars.ContextVar(
            "noreply", default=None
        )
        self._waitcontext: contextvars.ContextVar[tuple[int, int] | None] = contextvars.ContextVar(
            "wait", default=None
        )
        self._waitaof_context: contextvars.ContextVar[tuple[int, int, int] | None] = (
            contextvars.ContextVar("waitaof", default=None)
        )
        self.retry_policy = retry_policy
        self.type_adapter = type_adapter or TypeAdapter()

    def create_request(
        self,
        name: bytes,
        *arguments: ValueT,
        callback: Callable[..., T_co],
        execution_parameters: ExecutionParameters | None = None,
    ) -> CommandRequest[T_co]:
        """
        Factory method to create a command request awaitable.
        Subclasses of :class:`coredis.client.Client` can override this method
        if custom behavior is required. See :class:`~coredis.commands.CommandRequest`
        for details.

        :param name: The name of the command
        :param arguments: all arguments sent to the command
        :param callback: a callback that takes the RESP response and converts it
         into a shape to be returned
        :return: An instance of a command request bound to this client.
        """
        return CommandRequest(
            self, name, *arguments, callback=callback, execution_parameters=execution_parameters
        )

    @property
    def noreply(self) -> bool:
        if not hasattr(self, "_noreplycontext"):
            return False
        ctx = self._noreplycontext.get()
        if ctx is not None:
            return ctx
        return self.__noreply

    @property
    def requires_wait(self) -> bool:
        if not hasattr(self, "_waitcontext") or not self._waitcontext.get():
            return False
        return True

    @property
    def requires_waitaof(self) -> bool:
        if not hasattr(self, "_waitaof_context") or not self._waitaof_context.get():
            return False
        return True

    def _ensure_server_version(self, version: str | None) -> None:
        if self.verify_version and not Config.optimized and not self.server_version and version:
            try:
                self.server_version = Version(nativestr(version))
            except InvalidVersion:
                warnings.warn(
                    (
                        f"Server reported an invalid version: {version}."
                        "If this is expected you can dismiss this warning by passing "
                        "verify_version=False to the client constructor"
                    ),
                    category=UserWarning,
                )
                self.verify_version = False
                self.server_version = None

    async def _ensure_wait_and_persist(
        self, command: RedisCommandP, connection: BaseConnection
    ) -> None:
        wait = self._waitcontext.get()
        waitaof = self._waitaof_context.get()
        wait_request = None
        aof_request = None
        if wait and wait[0] > 0:
            wait_request = connection.create_request(CommandName.WAIT, *wait, decode=False)
        if waitaof and waitaof[0] > 0:
            aof_request = connection.create_request(CommandName.WAITAOF, *waitaof, decode=False)
        if wait_request and wait:
            wait_result = await wait_request
            if not cast(int, wait_result) >= wait[0]:
                raise ReplicationError(command.name, wait[0], wait[1])
        if aof_request and waitaof:
            aof_result = cast(tuple[int, int], await aof_request)
            if not (aof_result[0] >= waitaof[0] and aof_result[1] >= waitaof[1]):
                raise PersistenceError(command.name, *waitaof)

    def __repr__(self) -> str:
        return f"{type(self).__name__}<{self.connection_pool.location}>"

    async def scan_iter(
        self,
        match: StringT | None = None,
        count: int | None = None,
        type_: StringT | None = None,
    ) -> AsyncIterator[AnyStr]:
        """
        Make an iterator using the SCAN command so that the client doesn't
        need to remember the cursor position.
        """
        cursor = None

        while cursor != 0:
            cursor, data = await self.scan(cursor=cursor, match=match, count=count, type_=type_)

            for item in data:
                yield item

    async def sscan_iter(
        self,
        key: KeyT,
        match: StringT | None = None,
        count: int | None = None,
    ) -> AsyncIterator[AnyStr]:
        """
        Make an iterator using the SSCAN command so that the client doesn't
        need to remember the cursor position.
        """
        cursor = None

        while cursor != 0:
            cursor, data = await self.sscan(key, cursor=cursor, match=match, count=count)

            for item in data:
                yield item

    @overload
    def hscan_iter(
        self,
        key: KeyT,
        match: StringT | None = ...,
        count: int | None = ...,
    ) -> AsyncGenerator[tuple[AnyStr, AnyStr], None]: ...
    @overload
    def hscan_iter(
        self,
        key: KeyT,
        match: StringT | None = ...,
        count: int | None = ...,
        *,
        novalues: Literal[True],
    ) -> AsyncGenerator[AnyStr, None]: ...

    async def hscan_iter(
        self,
        key: KeyT,
        match: StringT | None = None,
        count: int | None = None,
        novalues: Literal[True] | None = None,
    ) -> AsyncGenerator[tuple[AnyStr, AnyStr], None] | AsyncGenerator[AnyStr, None]:
        """
        Make an iterator using the HSCAN command so that the client doesn't
        need to remember the cursor position.
        """
        cursor: int | None = None
        while cursor != 0:
            # TODO: find a better way to narrow the return type from hscan
            if novalues:
                cursor, fields = await self.hscan(
                    key, cursor=cursor, match=match, count=count, novalues=novalues
                )
                for item in fields:
                    yield item
            else:
                cursor, data = await self.hscan(key, cursor=cursor, match=match, count=count)

                for pair in data.items():
                    yield pair

    async def zscan_iter(
        self,
        key: KeyT,
        match: StringT | None = None,
        count: int | None = None,
    ) -> AsyncIterator[ScoredMember]:
        """
        Make an iterator using the ZSCAN command so that the client doesn't
        need to remember the cursor position.
        """
        cursor = None

        while cursor != 0:
            cursor, data = await self.zscan(
                key,
                cursor=cursor,
                match=match,
                count=count,
            )

            for item in data:
                yield item

    def register_script(self, script: RedisValueT) -> Script[AnyStr]:
        """
        Registers a Lua :paramref:`script`

        :return: A :class:`coredis.commands.script.Script` instance that is
         callable and hides the complexity of dealing with scripts, keys, and
         shas.
        """
        return Script[AnyStr](self, script)  # type: ignore

    @versionadded(version="3.1.0")
    async def register_library(
        self, name: StringT, code: StringT, replace: bool = False
    ) -> Library[AnyStr]:
        """
        Register a new library

        :param name: name of the library
        :param code: raw code for the library
        :param replace: Whether to replace the library when intializing. If ``False``
         an exception will be raised if the library was already loaded in the target
         redis instance.
        """
        return await Library[AnyStr](self, name=name, code=code, replace=replace)

    @versionadded(version="3.1.0")
    async def get_library(self, name: StringT) -> Library[AnyStr]:
        """
        Fetch a pre registered library

        :param name: name of the library
        """
        return await Library[AnyStr](self, name)

    @contextlib.contextmanager
    def ignore_replies(self: ClientT) -> Iterator[ClientT]:
        """
        Context manager to run commands without waiting for a reply.

        Example::

            client = coredis.Redis()
            with client.ignore_replies():
                assert None == await client.set("fubar", 1), "noreply"
            assert True == await client.set("fubar", 1), "reply"
        """
        noreply_reset_token = self._noreplycontext.set(True)
        try:
            yield self
        finally:
            self._noreplycontext.reset(noreply_reset_token)

    @contextlib.contextmanager
    def ensure_replication(
        self: ClientT, replicas: int = 1, timeout_ms: int = 100
    ) -> Iterator[ClientT]:
        """
        Context manager to ensure that commands executed within the context
        are replicated to atleast :paramref:`replicas` within
        :paramref:`timeout_ms` milliseconds.

        Internally this uses `WAIT <https://redis.io/commands/wait>`_ after
        each command executed within the context

        :raises: :exc:`coredis.exceptions.ReplicationError`

        Example::

            client = coredis.RedisCluster("localhost", 7000)
            with client.ensure_replication(1, 20):
                await client.set("fubar", 1)

        """
        replication_reset_token = self._waitcontext.set((replicas, timeout_ms))
        try:
            yield self
        finally:
            self._waitcontext.reset(replication_reset_token)

    @versionadded(version="4.12.0")
    @contextlib.contextmanager
    def ensure_persistence(
        self: ClientT,
        local: Literal[0, 1] = 0,
        replicas: int = 0,
        timeout_ms: int = 100,
    ) -> Iterator[ClientT]:
        """
        Context manager to ensure that commands executed within the context
        are synced to the AOF of a :paramref:`local` host and/or :paramref:`replicas`
        within :paramref:`timeout_ms` milliseconds.

        Internally this uses `WAITAOF <https://redis.io/commands/waitaof>`_ after
        each command executed within the context

        :raises: :exc:`coredis.exceptions.PersistenceError`

        Example for standalone client::

            client = coredis.Redis()
            with client.ensure_persistence(1, 0, 20):
                await client.set("fubar", 1)

        Example for cluster::

            client = coredis.RedisCluster("localhost", 7000)
            with client.ensure_persistence(1, 1, 20):
                await client.set("fubar", 1)

        """
        persistence_reset_token = self._waitaof_context.set((local, replicas, timeout_ms))
        try:
            yield self
        finally:
            self._waitaof_context.reset(persistence_reset_token)

    def should_quick_release(self, command: RedisCommandP) -> bool:
        return CommandFlag.BLOCKING not in COMMAND_FLAGS[command.name]


class Redis(Client[AnyStr]):
    connection_pool: ConnectionPool[Any]

    @overload
    def __init__(
        self: Redis[bytes],
        host: str | None = ...,
        port: int | None = ...,
        db: int = ...,
        *,
        username: str | None = ...,
        password: str | None = ...,
        credential_provider: AbstractCredentialProvider | None = ...,
        stream_timeout: float | None = ...,
        connect_timeout: float | None = ...,
        pool_timeout: float | None = ...,
        connection_pool: ConnectionPool[Any] | None = ...,
        connection_pool_cls: type[ConnectionPool[Any]] = ...,
        unix_socket_path: str | None = ...,
        encoding: str = ...,
        decode_responses: Literal[False] = ...,
        ssl: bool = ...,
        ssl_context: SSLContext | None = ...,
        ssl_keyfile: str | None = ...,
        ssl_certfile: str | None = ...,
        ssl_cert_reqs: Literal["optional", "required", "none"] | None = ...,
        ssl_check_hostname: bool | None = ...,
        ssl_ca_certs: str | None = ...,
        max_connections: int | None = ...,
        max_idle_time: int | None = ...,
        client_name: str | None = ...,
        verify_version: bool = ...,
        cache: AbstractCache | None = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        type_adapter: TypeAdapter | None = ...,
    ) -> None: ...

    @overload
    def __init__(
        self: Redis[str],
        host: str | None = ...,
        port: int | None = ...,
        db: int = ...,
        *,
        username: str | None = ...,
        password: str | None = ...,
        credential_provider: AbstractCredentialProvider | None = ...,
        stream_timeout: float | None = ...,
        connect_timeout: float | None = ...,
        pool_timeout: float | None = ...,
        connection_pool: ConnectionPool[Any] | None = ...,
        connection_pool_cls: type[ConnectionPool[Any]] = ...,
        unix_socket_path: str | None = ...,
        encoding: str = ...,
        decode_responses: Literal[True] = ...,
        ssl: bool = ...,
        ssl_context: SSLContext | None = ...,
        ssl_keyfile: str | None = ...,
        ssl_certfile: str | None = ...,
        ssl_cert_reqs: Literal["optional", "required", "none"] | None = ...,
        ssl_check_hostname: bool | None = ...,
        ssl_ca_certs: str | None = ...,
        max_connections: int | None = ...,
        max_idle_time: int | None = ...,
        client_name: str | None = ...,
        verify_version: bool = ...,
        cache: AbstractCache | None = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        type_adapter: TypeAdapter | None = ...,
    ) -> None: ...
    @overload
    def __init__(
        self: Self,
        host: str | None = ...,
        port: int | None = ...,
        db: int = ...,
        *,
        username: str | None = ...,
        password: str | None = ...,
        credential_provider: AbstractCredentialProvider | None = ...,
        stream_timeout: float | None = ...,
        connect_timeout: float | None = ...,
        pool_timeout: float | None = ...,
        connection_pool: ConnectionPool[Any] | None = ...,
        connection_pool_cls: type[ConnectionPool[Any]] = ...,
        unix_socket_path: str | None = ...,
        encoding: str = ...,
        decode_responses: bool = ...,
        ssl: bool = ...,
        ssl_context: SSLContext | None = ...,
        ssl_keyfile: str | None = ...,
        ssl_certfile: str | None = ...,
        ssl_cert_reqs: Literal["optional", "required", "none"] | None = ...,
        ssl_check_hostname: bool | None = ...,
        ssl_ca_certs: str | None = ...,
        max_connections: int | None = ...,
        max_idle_time: int | None = ...,
        client_name: str | None = ...,
        verify_version: bool = ...,
        cache: AbstractCache | None = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        type_adapter: TypeAdapter | None = ...,
    ) -> None: ...

    def __init__(
        self,
        host: str | None = "localhost",
        port: int | None = 6379,
        db: int = 0,
        *,
        username: str | None = None,
        password: str | None = None,
        credential_provider: AbstractCredentialProvider | None = None,
        stream_timeout: float | None = None,
        connect_timeout: float | None = None,
        pool_timeout: float | None = None,
        connection_pool: ConnectionPool[Any] | None = None,
        connection_pool_cls: type[ConnectionPool[Any]] = ConnectionPool,
        unix_socket_path: str | None = None,
        encoding: str = "utf-8",
        decode_responses: bool = False,
        ssl: bool = False,
        ssl_context: SSLContext | None = None,
        ssl_keyfile: str | None = None,
        ssl_certfile: str | None = None,
        ssl_cert_reqs: Literal["optional", "required", "none"] | None = None,
        ssl_check_hostname: bool | None = None,
        ssl_ca_certs: str | None = None,
        max_connections: int | None = None,
        max_idle_time: int | None = None,
        client_name: str | None = None,
        verify_version: bool = True,
        cache: AbstractCache | None = None,
        noreply: bool = False,
        noevict: bool = False,
        notouch: bool = False,
        retry_policy: RetryPolicy = ConstantRetryPolicy(
            (ConnectionError, TimeoutError), retries=2, delay=0.01
        ),
        type_adapter: TypeAdapter | None = None,
    ) -> None:
        """
        Changes

          - .. versionremoved:: 6.0.0

            - :paramref:`protocol_version` removed (and therefore support for RESP2)

          - .. versionchanged:: 6.0.0

            -  The client is now an async context manager and must always be used as such.

          - .. versionadded:: 4.12.0

            - :paramref:`retry_policy`
            - :paramref:`noevict`
            - :paramref:`notouch`
            - :meth:`Redis.ensure_persistence` context manager
            - Redis Module support

              - RedisJSON: :attr:`Redis.json`
              - RedisBloom:

                - BloomFilter: :attr:`Redis.bf`
                - CuckooFilter: :attr:`Redis.cf`
                - CountMinSketch: :attr:`Redis.cms`
                - TopK: :attr:`Redis.topk`
                - TDigest: :attr:`Redis.tdigest`
              - RedisTimeSeries: :attr:`Redis.timeseries`
              - RedisGraph: :attr:`Redis.graph`
              - RediSearch:

                - Search & Aggregation: :attr:`Redis.search`
                - Autocomplete: Added :attr:`Redis.autocomplete`

          - .. versionchanged:: 4.12.0

            - Removed :paramref:`retry_on_timeout` constructor argument. Use
              :paramref:`retry_policy` instead.

          - .. versionadded:: 4.3.0

            - Added :paramref:`connection_pool_cls`

          - .. versionchanged:: 4.0.0

            - :paramref:`non_atomic_cross_slot` defaults to ``True``
            - :paramref:`protocol_version`` defaults to ``3``

          - .. versionadded:: 3.11.0

            - Added :paramref:`noreply`

          - .. versionadded:: 3.9.0

            - If :paramref:`cache` is provided the client will check & populate
              the cache for read only commands and invalidate it for commands
              that could change the key(s) in the request.

          - .. versionchanged:: 3.5.0

            - The :paramref:`verify_version` parameter now defaults to ``True``

          - .. versionadded:: 3.1.0

            - The :paramref:`protocol_version` and :paramref:`verify_version`
              :parameters were added


        :param host: The hostname of the redis server
        :param port: The port at which th redis server is listening on
        :param db: database number to switch to upon connection
        :param username: Username for authenticating with the redis server
        :param password: Password for authenticating with the redis server
        :param credential_provider: CredentialProvider to get authentication credentials
        :param stream_timeout: Timeout (seconds) when reading responses from the server
        :param connect_timeout: Timeout (seconds) for establishing a connection to the server
        :param pool_timeout: Timeout (seconds) for acquiring a connection from the
         connection pool
        :param connection_pool: The connection pool instance to use. If not provided
         a new pool will be assigned to this client.
        :param connection_pool_cls: The connection pool class to use when constructing
         a connection pool for this instance.
        :param unix_socket_path: Path to the UDS which the redis server
         is listening at
        :param encoding: The codec to use to encode strings transmitted to redis
         and decode responses with. (See :ref:`handbook/encoding:encoding/decoding`)
        :param decode_responses: If ``True`` string responses from the server
         will be decoded using :paramref:`encoding` before being returned.
         (See :ref:`handbook/encoding:encoding/decoding`)
        :param ssl: Whether to use an SSL connection
        :param ssl_context: If provided the :class:`ssl.SSLContext` will be used when
         establishing the connection. Otherwise either the default context (if no other
         ssl related parameters are provided) or a custom context based on the other
         ``ssl_*`` parameters will be used.
        :param ssl_keyfile: Path of the private key to use
        :param ssl_certfile: Path to the certificate corresponding to :paramref:`ssl_keyfile`
        :param ssl_cert_reqs: Whether to try to verify the server's certificates and
         how to behave if verification fails (See :attr:`ssl.SSLContext.verify_mode`).
        :param ssl_check_hostname: Whether to enable hostname checking when establishing
         an ssl connection.
        :param ssl_ca_certs: Path to a concatenated certificate authority file or a directory
         containing several CA certifcates to use  for validating the server's certificates
         when :paramref:`ssl_cert_reqs` is not ``"none"``
         (See :meth:`ssl.SSLContext.load_verify_locations`).
        :param max_connections: Maximum capacity of the connection pool (Ignored if
         :paramref:`connection_pool` is not ``None``.
        :param max_idle_time: Maximum number of a seconds an unused connection is cached
         before it is disconnected.
        :param client_name: The client name to identifiy with the redis server
        :param verify_version: Validate redis server version against the documented
         version introduced before executing a command and raises a
         :exc:`CommandNotSupportedError` error if the required version is higher than
         the reported server version
        :param cache: If provided the cache will be used to avoid requests for read only
         commands if the client has already requested the data and it hasn't been invalidated.
         The cache is responsible for any mutations to the keys that happen outside of this client
        :param noreply: If ``True`` the client will not request a response for any
         commands sent to the server.
        :param noevict: Ensures that connections from the client will be excluded from the
         client eviction process even if we're above the configured client eviction threshold.
        :param notouch: Ensures that commands sent by the client will not alter the LRU/LFU of
         the keys they access.
        :param retry_policy: The retry policy to use when interacting with the redis server
        :param type_adapter: The adapter to use for serializing / deserializing customs types
         when interacting with redis commands.

        """
        if connection_pool and cache:
            raise RuntimeError("Parameters 'cache' and 'connection_pool' are mutually exclusive!")
        if not connection_pool:
            kwargs: ConnectionPoolParams[BaseConnection] = {
                "db": db,
                "username": username,
                "password": password,
                "credential_provider": credential_provider,
                "encoding": encoding,
                "stream_timeout": stream_timeout,
                "connect_timeout": connect_timeout,
                "timeout": pool_timeout,
                "max_connections": max_connections,
                "decode_responses": decode_responses,
                "max_idle_time": max_idle_time,
                "client_name": client_name,
                "noreply": noreply,
                "noevict": noevict,
                "notouch": notouch,
                "_cache": cache,
            }

            location: Location | None = None
            if unix_socket_path is not None:
                location = UnixDomainSocketLocation(unix_socket_path)
                kwargs["connection_class"] = UnixDomainSocketConnection
            else:
                if host is not None and port is not None:
                    location = TCPLocation(host, port)
                kwargs["connection_class"] = TCPConnection

                if ssl_context is not None:
                    kwargs["ssl_context"] = ssl_context
                elif ssl:
                    ssl_context = RedisSSLContext(
                        ssl_keyfile,
                        ssl_certfile,
                        ssl_cert_reqs,
                        ssl_ca_certs,
                        ssl_check_hostname,
                    ).get()
                    kwargs["ssl_context"] = ssl_context
            connection_pool = connection_pool_cls(location=location, **kwargs)
        super().__init__(
            connection_pool=connection_pool,
            verify_version=verify_version,
            noreply=noreply,
            retry_policy=retry_policy,
            type_adapter=type_adapter,
        )
        self._decodecontext: contextvars.ContextVar[bool | None,] = contextvars.ContextVar(
            "decode", default=None
        )
        self._encodingcontext: contextvars.ContextVar[str | None,] = contextvars.ContextVar(
            "decode", default=None
        )

    @classmethod
    @overload
    def from_url(
        cls,
        url: str,
        db: int | None = ...,
        *,
        decode_responses: Literal[False] = ...,
        verify_version: bool = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        cache: AbstractCache | None = ...,
        **kwargs: Any,
    ) -> Redis[bytes]: ...

    @classmethod
    @overload
    def from_url(
        cls,
        url: str,
        db: int | None = ...,
        *,
        decode_responses: Literal[True] = ...,
        verify_version: bool = ...,
        noreply: bool = ...,
        noevict: bool = ...,
        notouch: bool = ...,
        retry_policy: RetryPolicy = ...,
        cache: AbstractCache | None = ...,
        **kwargs: Any,
    ) -> Redis[str]: ...

    @classmethod
    def from_url(
        cls: type[RedisT],
        url: str,
        db: int | None = None,
        *,
        decode_responses: bool = False,
        verify_version: bool = True,
        noreply: bool = False,
        noevict: bool = False,
        notouch: bool = False,
        retry_policy: RetryPolicy = ConstantRetryPolicy(
            (ConnectionError, TimeoutError), retries=2, delay=0.01
        ),
        type_adapter: TypeAdapter | None = None,
        cache: AbstractCache | None = None,
        **kwargs: Any,
    ) -> RedisT:
        """
        Return a Redis client object configured from the given URL, which must
        use either the `redis:// scheme
        <http://www.iana.org/assignments/uri-schemes/prov/redis>`_ for RESP
        connections or the ``unix://`` scheme for Unix domain sockets.

        For example:

        - ``redis://[:password]@localhost:6379/0``
        - ``rediss://[:password]@localhost:6379/0``
        - ``unix://[:password]@/path/to/socket.sock?db=0``

        :paramref:`url` and :paramref:`kwargs` are passed as is to
        the :func:`coredis.pool.ConnectionPool.from_url`.
        """
        if decode_responses:
            return cls(
                decode_responses=True,
                verify_version=verify_version,
                noreply=noreply,
                retry_policy=retry_policy,
                type_adapter=type_adapter,
                connection_pool=ConnectionPool.from_url(
                    url,
                    db=db,
                    decode_responses=decode_responses,
                    noreply=noreply,
                    noevict=noevict,
                    notouch=notouch,
                    _cache=cache,
                    **kwargs,
                ),
            )
        else:
            return cls(
                decode_responses=False,
                verify_version=verify_version,
                noreply=noreply,
                retry_policy=retry_policy,
                type_adapter=type_adapter,
                connection_pool=ConnectionPool.from_url(
                    url,
                    db=db,
                    decode_responses=decode_responses,
                    noreply=noreply,
                    noevict=noevict,
                    notouch=notouch,
                    _cache=cache,
                    **kwargs,
                ),
            )

    @contextlib.asynccontextmanager
    async def __asynccontextmanager__(self) -> AsyncGenerator[Self]:
        async with self.connection_pool:
            yield self

    async def execute_command(
        self,
        command: RedisCommandP,
        callback: Callable[..., R] = NoopCallback(),
        **options: Unpack[ExecutionParameters],
    ) -> R:
        """
        Executes a command with configured retries and returns
        the parsed response
        """
        return await self.retry_policy.call_with_retries(
            lambda: self._execute_command(command, callback=callback, **options),
        )

    async def _execute_command(
        self,
        command: RedisCommandP,
        callback: Callable[..., R] = NoopCallback(),
        **options: Unpack[ExecutionParameters],
    ) -> R:
        pool = self.connection_pool
        quick_release = self.should_quick_release(command)
        should_block = not quick_release or self.requires_wait or self.requires_waitaof
        keys = KeySpec.extract_keys(command.name, *command.arguments)[0]
        cacheable = (
            command.name in CACHEABLE_COMMANDS
            and len(keys) == 1
            and not self.noreply
            and self._decodecontext.get() is None
        )
        cached_reply = None
        cache_hit = False
        use_cached = False
        reply = None
        released = False
        connection = await pool.get_connection()
        self._ensure_server_version(connection.server_version)
        try:
            if pool.cache and pool.cache.healthy:
                if connection.tracking_client_id != pool.cache.get_client_id(connection):
                    pool.cache.reset()
                    await connection.update_tracking_client(
                        True, pool.cache.get_client_id(connection)
                    )
                if command.name not in READONLY_COMMANDS:
                    pool.cache.invalidate(*keys)
                elif cacheable:
                    try:
                        cached_reply = cast(
                            R,
                            pool.cache.get(
                                command.name,
                                keys[0],
                                *command.arguments,
                            ),
                        )
                        use_cached = random.random() * 100.0 < min(100.0, pool.cache.confidence)
                        cache_hit = True
                    except KeyError:
                        pass
            if not (use_cached and cached_reply):
                request = connection.create_request(
                    command.name,
                    *command.arguments,
                    noreply=self.noreply,
                    decode=options.get("decode", self._decodecontext.get()),
                    encoding=self._encodingcontext.get(),
                    disconnect_on_cancellation=should_block,
                )
                # TODO: Fix this! using both the release & should_block
                #  flags to decide release logic is fragile. We should be
                #  releasing early even in the cached response flow.

                # if not blocking, no need to wait for reply
                if not should_block:
                    released = True
                    pool.release(connection)
                reply = await request
                await self._ensure_wait_and_persist(command, connection)
                if self.noreply:
                    return None  # type: ignore
            if pool.cache and cacheable:
                if cache_hit and not use_cached:
                    pool.cache.feedback(
                        command.name, keys[0], *command.arguments, match=cached_reply == reply
                    )
                if not cache_hit:
                    pool.cache.put(
                        command.name,
                        keys[0],
                        *command.arguments,
                        value=reply,
                    )
            return callback(cached_reply if cache_hit else reply)
        finally:
            if not released:
                pool.release(connection)

    @overload
    def decoding(
        self, mode: Literal[False], encoding: str | None = None
    ) -> contextlib.AbstractContextManager[Redis[bytes]]: ...

    @overload
    def decoding(
        self, mode: Literal[True], encoding: str | None = None
    ) -> contextlib.AbstractContextManager[Redis[str]]: ...

    @contextlib.contextmanager
    @versionadded(version="4.8.0")
    def decoding(self, mode: bool, encoding: str | None = None) -> Iterator[Redis[Any]]:
        """
        Context manager to temporarily change the decoding behavior
        of the client

        :param mode: Whether to decode or not
        :param encoding: Optional encoding to use if decoding. If not provided
         the :paramref:`~coredis.Redis.encoding` parameter provided to the client will
         be used.

        Example::

            client = coredis.Redis(decode_responses=True)
            await client.set("fubar", "baz")
            assert await client.get("fubar") == "baz"
            with client.decoding(False):
                assert await client.get("fubar") == b"baz"
                with client.decoding(True):
                    assert await client.get("fubar") == "baz"

        """
        decode_token = self._decodecontext.set(mode)
        encode_token = self._encodingcontext.set(encoding)
        try:
            yield self
        finally:
            self._decodecontext.reset(decode_token)
            self._encodingcontext.reset(encode_token)

    def pubsub(
        self,
        *,
        channels: Parameters[StringT] | None = None,
        channel_handlers: Mapping[StringT, SubscriptionCallback] | None = None,
        patterns: Parameters[StringT] | None = None,
        pattern_handlers: Mapping[StringT, SubscriptionCallback] | None = None,
        ignore_subscribe_messages: bool = False,
        retry_policy: RetryPolicy | None = CompositeRetryPolicy(
            ExponentialBackoffRetryPolicy(
                (ConnectionError,), retries=None, base_delay=0.1, max_delay=16, jitter=True
            ),
            ConstantRetryPolicy((TimeoutError,), retries=2, delay=0.1),
        ),
        subscription_timeout: float = 1,
        max_idle_seconds: float = 15,
    ) -> PubSub[AnyStr]:
        """
        Return a Pub/Sub instance that can be used to subscribe to channels
        and patterns and receive messages that get published to them.

        :param channels: channels that the constructed Pubsub instance should
         automatically subscribe to
        :param channel_handlers: Mapping of channels to automatically subscribe to
         and the associated handlers that will be invoked when a message is received
         on the specific channel.
        :param patterns: patterns that the constructed Pubsub instance should
         automatically subscribe to
        :param pattern_handlers: Mapping of patterns to automatically subscribe to
         and the associated handlers that will be invoked when a message is received
         on channel matching the pattern.
        :param ignore_subscribe_messages: Whether to skip subscription
         acknowledgement messages
        :param retry_policy: An explicit retry policy to use in the subscriber.
        :param subscription_timeout: Maximum amount of time in seconds to wait for
         acknowledgement of subscriptions.
        :param max_idle_seconds: Maximum duration (in seconds) to tolerate no
         messages from the server before performing a keepalive check with a
         ``PING``.
        """

        return PubSub[AnyStr](
            self.connection_pool,
            ignore_subscribe_messages=ignore_subscribe_messages,
            retry_policy=retry_policy,
            channels=channels,
            channel_handlers=channel_handlers,
            patterns=patterns,
            pattern_handlers=pattern_handlers,
            subscription_timeout=subscription_timeout,
            max_idle_seconds=max_idle_seconds,
        )

    def pipeline(
        self,
        transaction: bool = True,
        *,
        raise_on_error: bool = True,
        timeout: float | None = None,
    ) -> coredis.patterns.pipeline.Pipeline[AnyStr]:
        """
        Returns a new pipeline object that can queue multiple commands for
        batch execution.

        :param transaction: indicates whether all commands should be executed atomically.
        :param raise_on_error: Whether to raise errors upon executing the pipeline.
         If set to `False` errors will be accumulated and retrievable from the individual
         commands that had errors.
        :param timeout: If specified this value will take precedence over
         :paramref:`Redis.stream_timeout`
        """
        from coredis.patterns.pipeline import Pipeline

        return Pipeline[AnyStr](self, transaction, raise_on_error, timeout)

    def lock(
        self,
        name: StringT,
        timeout: float | None = None,
        sleep: float = 0.1,
        blocking: bool = True,
        blocking_timeout: float | None = None,
    ) -> Lock[AnyStr]:
        """
        Return a lock instance which can be used to guard resource access across
        multiple clients.

        :param name: key for the lock
        :param timeout: indicates a maximum life for the lock.
         By default, it will remain locked until :meth:`release` is called.
         ``timeout`` can be specified as a float or integer, both representing
         the number of seconds to wait.

        :param sleep: indicates the amount of time to sleep per loop iteration
         when the lock is in blocking mode and another client is currently
         holding the lock.

        :param blocking: indicates whether calling :meth:`acquire` should block until
         the lock has been acquired or to fail immediately, causing :meth:`acquire`
         to return ``False`` and the lock not being acquired. Defaults to ``True``.

        :param blocking_timeout: indicates the maximum amount of time in seconds to
         spend trying to acquire the lock. A value of ``None`` indicates
         continue trying forever. ``blocking_timeout`` can be specified as a
         :class:`float` or :class:`int`, both representing the number of seconds to wait.
        """
        from coredis.patterns.lock import Lock

        return Lock(self, name, timeout, sleep, blocking, blocking_timeout)

    async def transaction(
        self,
        func: Callable[[coredis.patterns.pipeline.Pipeline[AnyStr]], Awaitable[R]],
        *watches: KeyT,
        watch_delay: float | None = None,
    ) -> R:
        """
        Convenience method for executing the callable :paramref:`func` as a
        transaction while watching all keys specified in :paramref:`watches`.

        :param func: callable should expect a single argument which is a
         :class:`coredis.patterns.pipeline.Pipeline` object retrieved by calling
         :meth:`~coredis.Redis.pipeline`.
        :param watches: The keys to watch during the transaction
        :param watch_delay: Time in seconds to wait after each watch error before retrying
        """
        msg = "Caught WatchError in transaction, retrying..."
        while True:
            with catch({WatchError: lambda _: logger.warning(msg)}):
                async with self.pipeline(transaction=False) as pipe:
                    async with pipe.watch(*watches):
                        return await func(pipe)
            if watch_delay:
                await sleep(watch_delay)

    @overload
    def xconsumer(
        self,
        streams: Parameters[KeyT],
        *,
        buffer_size: int = ...,
        timeout: int | None = ...,
        group: Literal[None] = ...,
        consumer: Literal[None] = ...,
        auto_create: bool = ...,
        auto_acknowledge: bool = ...,
        start_from_backlog: bool = ...,
        **stream_parameters: StreamParameters,
    ) -> Consumer[AnyStr]: ...
    @overload
    def xconsumer(
        self,
        streams: Parameters[KeyT],
        *,
        buffer_size: int = ...,
        timeout: int | None = ...,
        group: StringT = ...,
        consumer: StringT = ...,
        auto_create: bool = ...,
        auto_acknowledge: bool = ...,
        start_from_backlog: bool = ...,
        **stream_parameters: StreamParameters,
    ) -> GroupConsumer[AnyStr]: ...
    @versionadded(version="6.0.0")
    @mutually_inclusive_parameters("group", "consumer")
    def xconsumer(
        self,
        streams: Parameters[KeyT],
        *,
        buffer_size: int = 0,
        timeout: int | None = None,
        group: StringT | None = None,
        consumer: StringT | None = None,
        auto_create: bool = True,
        auto_acknowledge: bool = False,
        start_from_backlog: bool = False,
        **stream_parameters: StreamParameters,
    ) -> Consumer[AnyStr] | GroupConsumer[AnyStr]:
        """
        Create a stream consumer for one or more Redis streams.

        Depending on whether ``group`` and ``consumer`` are provided, this method
        creates either a standalone stream consumer or a member of a stream
        consumer group.

        If ``group`` and ``consumer`` are not provided, a standalone stream
        consumer is created that starts reading from the latest entry of each
        stream provided in :paramref:`streams`.

        The latest entry is determined by calling
        :meth:`~coredis.Redis.xinfo_stream` and using the :data:`last-entry`
        attribute at the point of initializing the consumer instance or on first
        fetch (whichever comes first). If the stream(s) do not exist at the time
        of consumer creation, the consumer will simply start from the minimum
        identifier (``0-0``).

        If ``group`` and ``consumer`` are provided, a member of a stream consumer
        group is created. The consumer has an identical interface as
        :class:`coredis.patterns.streams.Consumer`.

        :param streams: the stream identifiers to consume from
        :param buffer_size: Size of buffer (per stream) to maintain. This
         translates to the maximum number of stream entries that are fetched
         on each request to redis.
        :param timeout: Maximum amount of time in milliseconds to block for new
         entries to appear on the streams the consumer is reading from.
        :param group: The name of the group this consumer is part of
        :param consumer: The unique name (within :paramref:`group`) of the consumer
        :param auto_create: If True the group will be created upon initialization
         or first fetch if it doesn't already exist.
        :param auto_acknowledge: If ``True`` the stream entries fetched will be fetched
         without needing to be acknowledged with :meth:`~coredis.Redis.xack` to remove
         them from the pending entries list.
        :param start_from_backlog: If ``True`` the consumer will start by fetching any pending
         entries from the pending entry list before considering any new messages
         not seen by any other consumer in the :paramref:`group`
        :param stream_parameters: Mapping of optional parameters to use
         by stream for the streams provided in :paramref:`streams`.

         .. warning:: Providing an ``identifier`` in ``stream_parameters`` has a different
            meaning for a group consumer. If the value is any valid identifier other than ``>``
            the consumer will only access the history of pending messages. That is, the set of
            messages that were delivered to this consumer (identified by :paramref:`consumer`)
            and never acknowledged.
        """
        from coredis.patterns.streams import Consumer, GroupConsumer

        if group is not None and consumer is not None:
            return GroupConsumer(
                self,
                streams,
                group=group,
                consumer=consumer,
                buffer_size=buffer_size,
                auto_create=auto_create,
                auto_acknowledge=auto_acknowledge,
                start_from_backlog=start_from_backlog,
                timeout=timeout,
                **stream_parameters,
            )
        else:
            return Consumer(
                self, streams, buffer_size=buffer_size, timeout=timeout, **stream_parameters
            )
